from dataclasses import field, dataclass
from typing import Any, Optional, Mapping, Union

from databricks.sdk.client_types import ClientType
from databricks.sdk.errors import ResourceDoesNotExist
from databricks.sdk.service.iam import Group as GroupV1, ComplexValue, User as UserV1
from databricks.sdk.service.iamv2 import Group as GroupV2, User as UserV2

from .service import IAMGroups, IAMUsers
from ..client import DatabricksResource

__all__ = [
    "IAMGroup",
    "IAMGroupLike",
    "IAMUser",
    "IAMUserLike"
]


IAMUserLike = Union[str, "IAMUser", UserV1, UserV2, ComplexValue, Mapping[str, Any]]
IAMGroupLike = Union[str, "IAMGroup", GroupV1, GroupV2, ComplexValue, Mapping[str, Any]]


@dataclass
class IAMGroup(DatabricksResource):
    service: IAMGroups = field(
        default_factory=IAMGroups.current,
        repr=False,
        compare=False
    )
    id: Optional[str] = None
    name: Optional[str] = None
    account_id: Optional[str] = None
    external_id: Optional[str] = None
    client_type: Optional[ClientType] = None
    entitlements: Optional[list[str]] = None
    members: Optional[list["IAMUser"]] = None

    def __str__(self):
        return self.name

    def __post_init__(self):
        if self.service is None:
            object.__setattr__(self, "service", IAMGroups.current())

        super().__post_init__()

    @classmethod
    def parse(
        cls,
        obj: Any,
        *,
        service: IAMGroups | None = None,
        client_type: Optional[ClientType] = None,
    ):
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, (GroupV1, GroupV2, ComplexValue)):
            return IAMGroup(
                service=service,
                client_type=client_type
            ).set_details(obj)

        elif isinstance(obj, Mapping):
            return cls.parse_mapping(obj, service=service, client_type=client_type)

        else:
            raise ValueError(f"Unsupported object type for parsing IAMGroup: {type(obj)}")

    @classmethod
    def parse_mapping(
        cls,
        data: Mapping[str, Any],
        *,
        service: IAMGroups | None = None,
        client_type: Optional[ClientType] = None,
        **kwargs
    ):
        if data:
            kwargs.update(data)

        group_id = kwargs.get("id") or kwargs.get("internal_id") or kwargs.get("value")
        display_name = kwargs.get("display_name") or kwargs.get("name") or kwargs.get("display")
        external_id = kwargs.get("external_id")
        account_id = kwargs.get("account_id")
        members = kwargs.get("members") or kwargs.get("users") or []

        if members:
            members = [
                IAMUser.parse(member, service=service.users)
                for member in members
                if member
            ]

        group = cls(
            service=service,
            client_type=client_type,
            id=group_id,
            name=display_name,
            external_id=external_id,
            account_id=account_id,
            members=members
        )

        return group

    def add_member(
        self,
        user: IAMUserLike,
        *,
        commit: bool = True
    ) -> "IAMGroup":
        user = IAMUser.parse(user, service=self.service.users)

        if user.id:
            if any(member.id == user.id for member in self.members):
                return self
        elif user.name:
            if any(member.name == user.name for member in self.members):
                return self
        elif user.username:
            if any(member.username == user.username for member in self.members):
                return self
        else:
            user = user.sync()

            if any(member.id == user.id for member in self.members):
                return self

        self.members.append(user)

        if commit:
            self.sync()

        return self

    def sync(self) -> "IAMGroup":
        assert self.id, "Group must have an ID to be committed"

        if self.client_type == ClientType.WORKSPACE:
            details = self.client.workspace_client().groups.update(
                id=self.id,
                display_name=self.name,
                external_id=self.external_id or None,
                members=[
                    member.complex_value
                    for member in self.members
                ] if self.members else None
            )

        elif self.client_type == ClientType.ACCOUNT:
            details = self.client.account_client().groups.update(
                id=self.id,
                display_name=self.name,
                external_id=self.external_id or None,
                members=[member.complex_value for member in self.members] if self.members else None
            )

        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")

        return self.set_details(details)

    def set_details(self, details: Union[GroupV1, GroupV2, "IAMGroup"]) -> "IAMGroup":
        if isinstance(details, GroupV1):
            self.id = details.id
            self.name = details.display_name
            self.external_id = details.external_id
            self.account_id = self.client.config.account_id
            self.entitlements = [_.value for _ in details.entitlements or ()]

            if details.meta:
                if details.meta.resource_type:
                    self.client_type = ClientType.ACCOUNT if details.meta.resource_type == "Group" else ClientType.WORKSPACE

            self.members = [
                IAMUser.parse(
                    member,
                    service=self.service.users,
                )
                for member in details.members or []
            ]

        elif isinstance(details, GroupV2):
            self.id = details.internal_id
            self.name = details.group_name
            self.external_id = details.external_id
            self.account_id = details.account_id
            self.members = None
            self.entitlements = None

        elif isinstance(details, ComplexValue):
            self.id = details.value
            self.name = details.display
            self.external_id = None
            self.account_id = None
            self.members = None
            self.entitlements = None

        elif isinstance(details, IAMGroup):
            self.id = details.id
            self.name = details.name
            self.external_id = details.external_id
            self.account_id = details.account_id
            self.client_type = details.client_type
            self.entitlements = details.entitlements
            self.members = details.members

        else:
            raise ValueError(f"Unsupported group details type: {type(details)}")

        return self


@dataclass
class IAMUser(DatabricksResource):
    service: IAMUsers = field(
        default_factory=IAMUsers.current,
        repr=False,
        compare=False
    )
    id: Optional[str] = None
    name: Optional[str] = None
    username: Optional[str] = None
    emails: Optional[list[str]] = None
    external_id: Optional[str] = None
    active: bool = True
    client_type: ClientType = ClientType.ACCOUNT

    def __post_init__(self):
        if self.service is None:
            object.__setattr__(self, "service", IAMUsers.current())

        super().__post_init__()

    def __str__(self):
        return self.name

    @classmethod
    def databricks_runtime(cls):
        return cls(
            id="databricks-runtime",
            name="Databricks Runtime",
            username="databricks-runtime",
            client_type=ClientType.ACCOUNT,
            active=True
        )

    @classmethod
    def parse(
        cls,
        obj: Any,
        *,
        service: IAMUsers | None = None,
        client_type: Optional[ClientType] = None
    ):
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, (UserV1, UserV2, ComplexValue)):
            user = cls(service=service, client_type=client_type or ClientType.ACCOUNT)
            user.set_details(obj)
            return user

        elif isinstance(obj, Mapping):
            return cls.parse_mapping(obj, service=service, client_type=client_type)

        elif isinstance(obj, str):
            return cls.parse_str(obj, service=service, client_type=client_type)

        else:
            raise ValueError(f"Unsupported object type for parsing IAMUser: {type(obj)}")

    @classmethod
    def parse_str(
        cls,
        value: str,
        *,
        service: IAMUsers | None = None,
        client_type: Optional[ClientType] = None
    ):
        if not value:
            raise ValueError("Value cannot be empty for parsing IAMUser from string")

        if "@" in value:
            emails = [value]
        else:
            emails = None

        return cls(
            service=service,
            id=None,
            name=None,
            emails=emails,
            client_type=client_type or ClientType.ACCOUNT
        )

    @classmethod
    def parse_mapping(
        cls,
        data: Mapping[str, Any],
        *,
        service: IAMUsers | None = None,
        client_type: Optional[ClientType] = None,
        **kwargs
    ):
        if data:
            kwargs.update(data)

        group_id = kwargs.get("id") or kwargs.get("internal_id") or kwargs.get("value")
        name = kwargs.get("display_name") or kwargs.get("name") or kwargs.get("display")
        external_id = kwargs.get("external_id")

        user = cls(
            service=service,
            id=group_id,
            name=name,
            external_id=external_id,
            client_type=client_type or ClientType.ACCOUNT
        )

        return user

    def set_details(self, details: UserV1 | UserV2 | ComplexValue) -> "IAMUser":
        if isinstance(details, UserV1):
            self.id = details.id
            self.name = details.display_name
            self.external_id = details.external_id
            self.emails = [_.value for _ in details.emails] if details.emails else None
            self.active = True if details.active is None else details.active

        elif isinstance(details, UserV2):
            self.id = details.internal_id
            self.external_id = details.external_id
            self.name = None
            self.username = details.username
            self.active = True

            if "@" in details.username:
                self.emails = [details.username]
            else:
                self.emails = None

        elif isinstance(details, ComplexValue):
            self.id = details.value
            self.name = details.display
            self.username = details.display
            self.active = True

            if "@" in details.display:
                self.emails = [details.display]
            else:
                self.emails = None

        elif isinstance(details, IAMUser):
            self.id = details.id
            self.name = details.name
            self.username = details.username
            self.emails = details.emails
            self.external_id = details.external_id
            self.active = details.active

        else:
            raise ValueError(f"Unsupported user details type: {type(details)}")

        return self

    def sync(self) -> "IAMUser":
        if not self.id:
            if self.username:
                found = next(self.service.list(user_name=self.username), None)

                if found is None:
                    raise ResourceDoesNotExist(
                        f"User with username '{self.username}' not found for syncing"
                    )

                return self.set_details(details=found)
            else:
                raise ValueError("User must have an ID to be synced")

        if self.client_type == ClientType.WORKSPACE:
            details = self.client.workspace_client().users.get(id=self.id)
        elif self.client_type == ClientType.ACCOUNT:
            details = self.client.account_client().users.get(id=self.id)
        else:
            raise ValueError(f"Unsupported client type: {self.client_type}")

        return self.set_details(details)

    @property
    def complex_value(self) -> ComplexValue:
        if not self.id:
            self.sync()
        return ComplexValue(
            value=self.id,
            display=self.name,
            ref=f"Users/{self.id}" if self.id else None
        )



try:
    import pyspark
except ImportError:
    from yggdrasil.environ import runtime_import_module

    pyspark = runtime_import_module(module_name="pyspark", pip_name="pyspark", install=True)

pyspark_sql = pyspark.sql


__all__ = [
    "pyspark",
    "pyspark_sql"
]



"""Remote execution helpers for Databricks command contexts."""

import dataclasses as dc
import inspect
import logging
import os
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Optional, Any, Callable, Dict, Union, Literal, TypeVar, \
    Mapping

from databricks.sdk.errors import DatabricksError
from databricks.sdk.service.compute import Language
from yggdrasil.concurrent.threading import Job
from yggdrasil.dataclasses.expiring import ExpiringDict
from yggdrasil.dataclasses.waiting import WaitingConfigArg
from yggdrasil.io.headers import DEFAULT_HOSTNAME
from yggdrasil.io.url import URL

if TYPE_CHECKING:
    from .cluster import Cluster
    from .command_execution import CommandExecution


__all__ = [
    "ExecutionContext",
    "exclude_env_key"
]

EXCLUDED_ENV_EXACT_KEYS = frozenset([
    'ALLUSERSPROFILE', 'APPDATA',
    'ARM_CLIENT_ID', 'ARM_CLIENT_SECRET', 'ARM_ENVIRONMENT', 'ARM_RESOURCE_ID', 'ARM_TENANT_ID',
    'CLASSPATH', 'CLICOLOR', 'CLICOLOR_FORCE', 'CLUSTER_DB_HOME', 'COMMONPROGRAMFILES', 'COMMONPROGRAMFILES(X86)',
    'COMMONPROGRAMW6432', 'COMPUTERNAME', 'COMSPEC', 'DATABRICKS_CLUSTER_ID', 'DATABRICKS_CLUSTER_LIBS_PYTHON_ROOT_DIR',
    'DATABRICKS_CLUSTER_LIBS_ROOT_DIR', 'DATABRICKS_CLUSTER_LIBS_R_ROOT_DIR',
    'DATABRICKS_HOST', 'DATABRICKS_INSTANCE_ID', 'DATABRICKS_LIBS_NFS_ROOT_DIR',
    'DATABRICKS_LIBS_NFS_ROOT_PATH', 'DATABRICKS_ROOT_VIRTUALENV_ENV', 'DATABRICKS_RUNTIME_VERSION',
    'DATABRICKS_TOKEN', 'DATA_SECURITY_MODE', 'DBX_WORKSPACE_URL', 'DB_HOME', 'DEBUGINFOD_URLS',
    'DEFAULT_DATABRICKS_ROOT_VIRTUALENV_ENV', 'DEFAULT_PYTHON_ENVIRONMENT', 'DISABLE_LOCAL_FILESYSTEM',
    'DRIVERDATA', 'DRIVER_PID_FILE', 'DRIVER_REPL_ID', 'DRIVER_STARTUP_OBSERVABILITY_ENABLED',
    'EFC_12712_1262719628', 'EFC_12712_1592913036', 'EFC_12712_2283032206', 'EFC_12712_2775293581',
    'EFC_12712_3789132940', 'ENABLE_APPCDS', 'ENABLE_CLASSLOADING_LOGS', 'ENABLE_COMMAND_OUTPUT_TRUNCATION',
    'ENABLE_DRIVER_DEVELOPER_MODE', 'ENABLE_IPTABLES', 'ENABLE_KEEPALIVE_COMMAND_CONTEXT', 'ENABLE_REPL_LOGGING',
    'ENABLE_TRACEPARENT_REPL_PROPAGATION', 'FORCE_COLOR', 'GIT_PAGER', 'GRPC_GATEWAY_TOKEN',
    'HALT_VARIABLE_RESOLVE_THREADS_ON_STEP_RESUME', 'HF_DATASETS_CACHE', 'HIVE_HOME', 'HOME', 'HOMEDRIVE',
    'HOMEPATH', 'HTTPS_PROXY', 'HTTP_PROXY', 'ICU_DATA', 'IDE_PROJECT_ROOTS', 'IPYTHONENABLE', 'JAVA_HOME',
    'JAVA_OPTS', 'JUPYTER_WIDGETS_ECHO', 'KOALAS_USAGE_LOGGER', 'LANG', 'LIBRARY_ROOTS', 'LOCALAPPDATA',
    'LOGNAME', 'LOGONSERVER', 'MAIL', 'MASTER', 'MEDSITE', 'MLFLOW_CONDA_HOME', 'MLFLOW_DEPLOYMENTS_TARGET',
    'MLFLOW_GATEWAY_URI', 'MLFLOW_PYTHON_EXECUTABLE', 'MLFLOW_REGISTRY_URI', 'MLFLOW_TRACKING_URI', 'MPLBACKEND',
    'NEXTHINK', 'NO_PROXY', 'NUMBER_OF_PROCESSORS', 'OLDPWD', 'OMPI_MCA_btl_tcp_if_include', 'ONEDRIVE',
    'ONEDRIVECOMMERCIAL', 'OPENSSL_FORCE_FIPS_MODE', 'OS', 'PAGER', 'PATH', 'PATHEXT', 'PINNED_THREAD_MODE',
    'PIPELINE_UDS_CONNECT_MODE', 'PIP_NO_INPUT', 'PROCESSOR_ARCHITECTURE', 'PROCESSOR_IDENTIFIER', 'PROCESSOR_LEVEL',
    'PROCESSOR_REVISION', 'PROGRAMDATA', 'PROGRAMFILES', 'PROGRAMFILES(X86)', 'PROGRAMW6432', 'PROJ_DATA', 'PROMPT',
    'PS1', 'PSMODULEPATH', 'PUBLIC', 'PWD', 'PYCHARM_HELPERS_DIR', 'PYCHARM_HOSTED',
    'PYDEVD_DISABLE_FILE_VALIDATION', 'PYDEVD_INTERRUPT_THREAD_TIMEOUT', 'PYDEVD_LOAD_VALUES_ASYNC',
    'PYDEVD_USE_FRAME_EVAL', 'PYENV_ROOT', 'PYSPARK_GATEWAY_PORT', 'PYSPARK_GATEWAY_SECRET', 'PYSPARK_PYTHON',
    'PYTEST_CURRENT_TEST', 'PYTEST_RUN_CONFIG', 'PYTEST_VERSION', 'PYTHONHASHSEED', 'PYTHONIOENCODING',
    'PYTHONPATH', 'PYTHONUNBUFFERED', 'PYTHON_REPL_SAFE_CONFIG_MAP', 'RAY_TMPDIR', 'R_LIBS', 'SCALA_VERSION',
    'SESSIONNAME', 'SHELL', 'SHLVL', 'SPARK_AUTH_SOCKET_TIMEOUT', 'SPARK_BUFFER_SIZE', 'SPARK_CONF_DIR',
    'SPARK_DIST_CLASSPATH', 'SPARK_ENV_LOADED', 'SPARK_HOME', 'SPARK_LOCAL_DIRS', 'SPARK_LOCAL_IP', 'SPARK_PUBLIC_DNS',
    'SPARK_SCALA_VERSION', 'SPARK_WORKER_MEMORY', 'SUDO_COMMAND', 'SUDO_GID', 'SUDO_UID', 'SUDO_USER', 'SYSTEMDRIVE',
    'SYSTEMROOT', 'TEAMCITY_VERSION', 'TEMP', 'TERM', 'TMP', 'TS_ADSITE_DRIVE', 'TS_APP32', 'TS_APP64', 'TS_APPRW',
    'TS_CONDA', 'TS_DFSN_ROOT', 'TS_IOADDIN', 'TS_PYTHON', 'TZDIR', 'UATDATA', 'USER', 'USERDNSDOMAIN', 'USERDOMAIN',
    'USERDOMAIN_ROAMINGPROFILE', 'USERNAME', 'USERPROFILE', 'USE_LOW_IMPACT_MONITORING', 'VIRTUAL_ENV',
    'VIRTUAL_ENV_PROMPT', 'WINDIR', 'ZES_ENABLE_SYSMAN', '_JB_PPRINT_PRIMITIVES', '_OLD_VIRTUAL_PATH',
    '_OLD_VIRTUAL_PROMPT', '_PIP_USE_IMPORTLIB_METADATA', '_RJEM_MALLOC_CONF', 'container'
])

EXCLUDED_ENV_PREFIXES = (
    "ARM_",
    "DATABRICKS_",
    "SPARK_",
    "PYSPARK_",
    "PYTHON_",
    "PYTEST_",
    "PYDEVD_",
    "PYCHARM_",
    "VSCODE_",
    "MLFLOW_",
    "TS_",
    "EFC_",
    "GIT_",
    "GITGUARDIAN_",
    "COMMONPROGRAMFILES",
    "FPS_BROWSER"
)

def exclude_env_key(key: str) -> bool:
    """Return True when an environment variable key should be excluded."""
    k = key.upper()

    return (
        k in EXCLUDED_ENV_EXACT_KEYS
        or k.startswith(EXCLUDED_ENV_PREFIXES)
    )


LOGGER = logging.getLogger(__name__)
UPLOADED_PACKAGE_ROOTS: Dict[str, ExpiringDict] = {}
BytesLike = Union[bytes, bytearray, memoryview]
F = TypeVar("F", bound=Callable[..., Any])

_CTX_RUNTIME_FIELDS = frozenset({"_lock",})
_CTX_RESET_FIELDS = frozenset({"_remote_metadata"})


@dc.dataclass
class RemoteMetadata:
    context_path: str
    tmp_path: str
    libs_path: str


def _bind_call_args(
    func: Callable,
    args: tuple,
    kwargs: dict,
) -> tuple[tuple, dict]:
    """Validate *args*/*kwargs* against *func*'s signature, fill in defaults.

    Parameters
    ----------
    func:
        The callable whose signature to inspect.
    args:
        Positional arguments intended for *func*.
    kwargs:
        Keyword arguments intended for *func*.

    Returns
    -------
    tuple[tuple, dict]
        ``(args, kwargs)`` with defaults for optional parameters filled in
        to the *kwargs* dict.

    Raises
    ------
    TypeError
        If required parameters are missing or unexpected keyword arguments
        are passed.
    """
    try:
        sig = inspect.signature(func)
    except (ValueError, TypeError):
        # Built-ins or C extensions may not expose a signature; skip validation.
        return args, kwargs

    try:
        bound = sig.bind(*args, **kwargs)
    except TypeError as exc:
        raise TypeError(
            f"Invalid arguments for {getattr(func, '__qualname__', repr(func))!r}: {exc}"
        ) from exc

    bound.apply_defaults()

    # Reconstruct positional and keyword arguments from the bound mapping.
    # Parameters that were originally positional stay positional; the rest
    # (keyword-only + defaults that were filled in) go into kwargs.
    new_args: list = []
    new_kwargs: dict = {}

    for param_name, param in sig.parameters.items():
        if param_name not in bound.arguments:
            continue
        value = bound.arguments[param_name]
        kind = param.kind
        if kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ):
            new_args.append(value)
        elif kind == inspect.Parameter.VAR_POSITIONAL:
            new_args.extend(value)
        elif kind in (
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.VAR_KEYWORD,
        ):
            if kind == inspect.Parameter.VAR_KEYWORD:
                new_kwargs.update(value)
            else:
                new_kwargs[param_name] = value

    return tuple(new_args), new_kwargs


@dc.dataclass
class ExecutionContext:
    """
    Lightweight wrapper around Databricks command execution context for a cluster.

    Can be used directly:

        ctx = ExecutionContext(cluster=my_cluster)
        ctx.open()
        ctx.execute("print(1)")
        ctx.close()

    Or as a context manager to reuse the same remote context for multiple commands:

        with ExecutionContext(cluster=my_cluster) as ctx:
            ctx.execute("x = 1")
            ctx.execute("print(x + 1)")
    """
    cluster: "Cluster"
    context_id: Optional[str] = None
    context_key: Optional[str] = None
    ephemeral: bool = dc.field(default=True, repr=False, compare=False, hash=False)
    language: Optional[Language] = dc.field(default=None, repr=False, compare=False, hash=False)

    _remote_metadata: Optional[RemoteMetadata] = dc.field(default=None, init=False, repr=False, compare=False, hash=False)
    _requirements: Optional[list[tuple[str]]] = dc.field(default=None, init=False, repr=False, compare=False, hash=False)
    _pyenv_check_timestamp: int = dc.field(default=0, init=False, repr=False, compare=False, hash=False)
    _lock: threading.RLock = dc.field(default_factory=threading.RLock, init=False, repr=False, compare=False, hash=False)

    def __getstate__(self) -> dict:
        """Serialize context state for pickling.

        Drops unpickable threading primitives and resets fields whose
        values are only meaningful in the originating process:

        - ``_lock``: RLock is not picklable and must always be reconstructed
        - ``_remote_metadata``: contains a ``temp_path`` that only exists on
          the remote cluster's filesystem; stale after transport.  The
          ``site_packages_path`` and ``os_env`` within it are also
          process/host-specific.  Drop it and let the lazy property
          re-fetch on first use.

        ``_uploaded_package_roots`` is preserved: remote paths remain valid
        across processes as long as the cluster session is alive, so we avoid
        redundant re-uploads.

        Returns:
            A compact, pickle-ready state dictionary.
        """
        state = {}

        for key, value in self.__dict__.items():
            if key in _CTX_RUNTIME_FIELDS:
                continue
            if key in _CTX_RESET_FIELDS:
                state[key] = None  # preserve key for attribute completeness
                continue
            state[key] = value

        return state

    def __setstate__(self, state: dict) -> None:
        """Restore context state after unpickling.

        Always constructs a fresh RLock — never attempts to restore a
        serialized one.  Ensures all expected attributes are present even
        when the state was produced by an older serialized form.

        Args:
            state: Serialized state dictionary.
        """
        state["_lock"] = threading.RLock()  # always fresh

        self.__dict__.update(state)

    def __enter__(self) -> "ExecutionContext":
        """Enter a context manager, opening a remote execution context."""
        if self.context_id is None:
            return self.create(
                language=self.language,
                context_key=self.context_key
            )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager and close the remote context if created."""
        self.close(wait=False, raise_error=False)

    def __repr__(self):
        return "%s(url=%s)" % (
            self.__class__.__name__,
            self.url()
        )

    def __str__(self):
        return self.url().to_string()

    def url(self) -> URL:
        url = self.cluster.url()

        return url.with_query_items({
            "context": self.context_id or "unknown"
        })

    @property
    def client(self):
        return self.cluster.client

    @property
    def cluster_id(self):
        return self.cluster.cluster_id

    @property
    def remote_metadata(self) -> RemoteMetadata:
        """Fetch and cache remote environment metadata for the cluster."""
        # fast path (no lock)
        if self._remote_metadata is not None:
            return self._remote_metadata

        if not self.context_key:
            self.context_key = DEFAULT_HOSTNAME

        context_path = f"/local_disk0/.ephemeral_nfs/context/{self.context_key}"
        tmp_path = context_path + "/tmp/"
        libs_path = context_path + "/python/lib/site-packages"

        self._remote_metadata = RemoteMetadata(
            context_path=context_path,
            tmp_path=tmp_path,
            libs_path=libs_path
        )

        return self._remote_metadata

    @property
    def requirements(self):
        if self._requirements is not None:
            return self._requirements

        command = f"uv pip --directory {str(self.remote_metadata.libs_path)!r} list --format=json"

        try:
            reqs = self.command(
                command_str=command,
                language="shell",
            ).start().result()

            self._requirements = [
                (kw["name"], kw["version"])
                for kw in reqs
            ]
        except Exception as e:
            if "exit code 2" in str(e):
                self._requirements = []
            else:
                raise e

        return self._requirements

    # ------------ internal helpers ------------
    def create(
        self,
        *,
        language: "Language",
        context_key: Optional[str] = None,
        wait: WaitingConfigArg = True,
        ephemeral: bool = True
    ) -> "ExecutionContext":
        """Create a command execution context, retrying if needed.

        Args:
            language: The Databricks command language to use.
            context_key: Constant string key value
            wait: Waiting config to update

        Returns:
            The created command execution context response.
        """
        if self.context_id and self.language == language:
            return self

        client = self.client.workspace_client().command_execution

        try:
            with ThreadPoolExecutor(max_workers=1) as ex:
                fut = ex.submit(
                    client.create,
                    cluster_id=self.cluster_id,
                    language=language,
                )

                try:
                    created = fut.result(timeout=10).response
                except TimeoutError:
                    self.cluster.ensure_running(wait=True)

                    created = client.create(
                        cluster_id=self.cluster_id,
                        language=language,
                    ).response
        except Exception as e:
            LOGGER.warning(e)

            self.cluster.ensure_running(wait=True)

            created = client.create(
                cluster_id=self.cluster_id,
                language=language,
            ).response

        instance = ExecutionContext(
            cluster=self.cluster,
            context_id=created.id,
            context_key=context_key or self.context_key or os.urandom(8).hex(),
            language=language,
            ephemeral=ephemeral
        )

        return instance

    def connect(
        self,
        *,
        language: Optional[Language] = None,
        wait: WaitingConfigArg = True,
        reset: bool = False,
    ) -> "ExecutionContext":
        """Create a remote command execution context if not already open.

        Args:
            language: Optional language override for the context.
            wait: Wait config
            reset: Reset existing if connected

        Returns:
            The connected ExecutionContext instance.
        """
        if self.context_id is not None:
            if not reset:
                return self

            LOGGER.info(
                "%s reset connection",
                self
            )

            self.close(wait=False)

        language = language or self.language

        if language is None:
            language = Language.PYTHON

        return self.create(
            language=language,
            context_key=self.context_key,
            wait=wait
        )

    def close(
        self,
        wait: bool = True,
        raise_error: bool = True
    ) -> None:
        """Destroy the remote command execution context if it exists.

        Returns:
            None.
        """
        if not self.context_id:
            return

        client = self.client.workspace_client()

        try:
            if wait:
                client.command_execution.destroy(
                    cluster_id=self.cluster.cluster_id,
                    context_id=self.context_id,
                )
            else:
                Job.make(
                    client.command_execution.destroy,
                    cluster_id=self.cluster_id,
                    context_id=self.context_id
                ).fire_and_forget()
        except DatabricksError:
            if raise_error:
                raise
        finally:
            self.context_id = None

    # ------------ public API ------------
    def syspath_lines(self, environ: Optional[Mapping[str, str]] = None) -> str:
        """Build the preamble that sets ``sys.path`` and environment variables.

        The snippet is prepended to every Python command sent to the
        remote execution context.  It:

        1. Ensures the remote libs path is on ``sys.path``.
        2. Injects any extra environment variables into ``os.environ``.

        Parameters
        ----------
        environ:
            Optional mapping of environment variables to propagate.

        Returns
        -------
        str
            A multi-line Python snippet suitable for ``exec()``.
        """
        lines = f"""\
import base64, os, traceback, json, sys
_p = os.path.expanduser({self.remote_metadata.libs_path!r})
if _p not in sys.path:
    sys.path.insert(0, _p)"""

        if environ:
            lines += f"""
for _k, _v in {dict(environ)!r}.items():
    os.environ[_k] = str(_v)"""

        return lines

    def command(
        self,
        command: Optional[str | Callable] = None,
        *,
        command_str: Optional[str] = None,
        language: Optional[Language | Literal["python", "r", "sql", "scala", "shell"]] = None,
        context: Optional["ExecutionContext"] = None,
        command_id: Optional[str] = None,
        func: Optional[Callable] = None,
        environ: Optional[Mapping] = None
    ) -> "CommandExecution":
        from .command_execution import CommandExecution

        context = self if context is None else context

        out_environ = {
            k: v
            for k, v in os.environ.items()
            if not exclude_env_key(k)
        }

        if environ:
            if not isinstance(environ, Mapping):
                out_environ.update({
                    k: os.getenv(k)
                    for k in (str(_) for _ in environ if _)
                    if k and not exclude_env_key(k)
                })
            else:
                out_environ.update({
                    str(k): str(v)
                    for k, v in environ.items()
                    if not exclude_env_key(k)
                })

        if isinstance(command, str):
            command_str = command
        elif callable(command):
            func = command

        if language == "shell":
            language = Language.PYTHON
            command_str = f"""
import subprocess, sys, shlex, pathlib

cmd = shlex.split({str(command_str)!r})
cmd = [str(pathlib.Path(arg).expanduser()) if arg.startswith("~/") else arg for arg in cmd]

p = subprocess.run(cmd, text=True, capture_output=True)

if p.returncode != 0:
    raise RuntimeError(
        f"Command {{cmd}} failed with exit code {{p.returncode}}:\\n"
        f"stderr: {{p.stderr.strip()}}"
    )
"""
        elif isinstance(language, str):
            language = Language[language]

        if language == Language.PYTHON and command_str:
            command_str = self.syspath_lines(out_environ) + "\n" + command_str

        return CommandExecution(
            context=context,
            command_id=command_id,
            language=language,
            command=command_str,
            pyfunc=func,
            environ=out_environ
        )

    # Prefix used to signal that the payload literal in the remote
    # command is a Databricks path to read from, not an inline value.
    _DBXPATH_PREFIX = "DBXPATH:"

    # Conservative payload-size threshold.  Databricks command execution
    # has a ~1 MB command-text limit; we leave headroom for the preamble
    # and wrapper code.
    _MAX_INLINE_PAYLOAD_BYTES = 900_000

    def make_python_function_command(
        self,
        func: Callable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
        tag: str = "__CALL_RESULT__",
        environ: Optional[Mapping] = None,
    ) -> tuple[str, Optional[str]]:
        """Build a remote Python command that executes a serialised callable.

        The generated snippet:

        1. Prepends ``sys.path`` / environment preamble.
        2. Checks whether the payload literal starts with ``DBXPATH:``.
           If so, reads the base64 payload from the Databricks path;
           otherwise treats it as an inline base64 string.
        3. Deserialises ``[func, args, kwargs]`` from the payload.
        4. Calls ``func(*args, **kwargs)``.
        5. Serialises the result (or the exception) and prints it after
           the *tag* sentinel so that
           :meth:`CommandExecution.decode_response` can split logs from
           the return value.

        The ``try/except`` wrapper guarantees the tag is **always**
        printed, even when the function raises.  The caller-side
        ``result()`` method deserialises the exception and re-raises it.

        When the serialised payload exceeds
        :attr:`_MAX_INLINE_PAYLOAD_BYTES` the payload is uploaded to a
        temporary Databricks path (Volumes / DBFS) and the command
        references it via the ``DBXPATH:`` prefix.  The caller is
        responsible for deleting this path after use (see
        :meth:`CommandExecution.result`).

        Parameters
        ----------
        func:
            The callable to execute remotely.
        args:
            Positional arguments for *func*.
        kwargs:
            Keyword arguments for *func*.
        tag:
            Sentinel string separating stdout logs from the serialised
            return value.
        environ:
            Extra environment variables to inject on the remote side.

        Returns
        -------
        tuple[str, str | None]
            ``(command_string, remote_payload_path_or_None)``.
            *remote_payload_path* is the Databricks path where the
            payload was uploaded, or ``None`` when the payload was
            embedded inline.

        Raises
        ------
        TypeError
            When *args* / *kwargs* do not satisfy *func*'s signature
            (missing required arguments, unexpected keyword arguments, etc.).
        """
        from yggdrasil.pickle.ser import dumps

        args, kwargs = _bind_call_args(func, args or (), kwargs or {})

        payload: str = dumps([func, args, kwargs], b64=True)
        remote_payload_path: Optional[str] = None

        if len(payload) > self._MAX_INLINE_PAYLOAD_BYTES:
            tmp = self.client.tmp_path(
                suffix="-pyfunc", extension="b64",
                max_lifetime=3600,
            )
            payload_bytes = (
                payload.encode("ascii")
                if isinstance(payload, str)
                else payload
            )
            tmp.write_bytes(payload_bytes)
            remote_payload_path = str(tmp)

            # The remote snippet receives a DBXPATH:-prefixed literal
            # and reads the actual base64 data from the path.
            payload_literal = f"{self._DBXPATH_PREFIX}{remote_payload_path}"
        else:
            payload_literal = payload

        cmd = f"""\
{self.syspath_lines(environ)}

from yggdrasil.pickle.ser import loads, dumps
_DBXPATH_PREFIX = {self._DBXPATH_PREFIX!r}
_MAX_INLINE = {self._MAX_INLINE_PAYLOAD_BYTES!r}
_raw = {payload_literal!r}

if _raw.startswith(_DBXPATH_PREFIX):
    from yggdrasil.databricks.client import DatabricksClient as _DBC
    _client = _DBC.current()
    _path = _raw[len(_DBXPATH_PREFIX):]
    _raw = _client.dbfs_path(_path).read_bytes().decode("ascii")
_f, _a, _k = loads(_raw)
_r = dumps(_f(*_a, **_k), b64=True)

if len(_r) > _MAX_INLINE:
    from yggdrasil.databricks.client import DatabricksClient as _DBC
    _client = _DBC.current()
    _tmp = _client.tmp_path(suffix="-pyresult", extension="b64", max_lifetime=3600)
    _tmp.write_bytes(_r.encode("ascii") if isinstance(_r, str) else _r)
    _r = _DBXPATH_PREFIX + str(_tmp)
    
print({tag!r} + _r, flush=True)
"""
        return cmd, remote_payload_path

    def is_in_databricks_environment(self):
        """Return True when running on a Databricks runtime."""
        return self.cluster.client.is_in_databricks_environment()



from __future__ import annotations

from fastapi import FastAPI, Request
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import JSONResponse

from .config import Settings, get_settings
from .exceptions import register_exception_handlers
from .routers import excel_router, python_router, system_router
from .services import DatabricksExcelService, PythonService, SystemService


def create_app(settings: Settings | None = None) -> FastAPI:
    settings = settings or get_settings()

    app = FastAPI(
        title=settings.app_name,
        version=settings.app_version,
        docs_url=settings.docs_url,
        redoc_url=settings.redoc_url,
        openapi_url=settings.openapi_url,
    )

    app.state.settings = settings
    app.state.system_service = SystemService(settings)
    app.state.python_service = PythonService(settings)
    app.state.databricks_excel_service = DatabricksExcelService(settings)

    @app.middleware("http")
    async def local_only_middleware(request: Request, call_next):
        if settings.allow_remote:
            return await call_next(request)

        client_host = request.client.host if request.client else None
        if client_host and client_host not in settings.local_clients:
            return JSONResponse(
                status_code=403,
                content={
                    "detail": (
                        "Remote access is disabled. Bind locally or set "
                        "YGG_FASTAPI_ALLOW_REMOTE=1 to allow non-local clients."
                    )
                },
            )

        return await call_next(request)

    register_exception_handlers(app)
    app.add_middleware(GZipMiddleware, minimum_size=1024, compresslevel=5)

    app.include_router(
        system_router,
        prefix=f"{settings.api_prefix}{settings.system_prefix}",
    )
    app.include_router(
        python_router,
        prefix=f"{settings.api_prefix}{settings.python_prefix}",
    )
    app.include_router(
        excel_router,
        prefix=f"{settings.api_prefix}{settings.python_prefix}{settings.excel_prefix}",
    )

    return app


app = create_app()



from __future__ import annotations

from array import array
from collections import deque
from collections.abc import Generator as AbcGenerator
from collections.abc import Iterator as AbcIterator
from collections.abc import Mapping as AbcMapping
from dataclasses import dataclass
from typing import ClassVar, Generic, Iterator, Mapping

from yggdrasil.io import BytesIO
from yggdrasil.pickle.ser.constants import CODEC_NONE
from yggdrasil.pickle.ser.serialized import Serialized, T
from yggdrasil.pickle.ser.tags import Tags

__all__ = [
    "CollectionSerialized",
    "LargeCollectionSerialized",
    "ArraySerialized",
    "LargeArraySerialized",
    "ListSerialized",
    "LargeListSerialized",
    "TupleSerialized",
    "LargeTupleSerialized",
    "SetSerialized",
    "LargeSetSerialized",
    "FrozenSetSerialized",
    "LargeFrozenSetSerialized",
    "DequeSerialized",
    "LargeDequeSerialized",
    "MappingSerialized",
    "LargeMappingSerialized",
    "GeneratorSerialized",
    "LargeGeneratorSerialized",
    "IteratorSerialized",
    "LargeIteratorSerialized",
]

_U32_MAX = 0xFFFFFFFF


# ============================================================================
# low-level count helpers
# ============================================================================

def _read_u32(buffer: BytesIO) -> int:
    raw = buffer.read(4)
    if len(raw) != 4:
        raise ValueError(f"Expected 4 bytes for u32, got {len(raw)}")
    return int.from_bytes(raw, "big", signed=False)


def _read_u64(buffer: BytesIO) -> int:
    raw = buffer.read(8)
    if len(raw) != 8:
        raise ValueError(f"Expected 8 bytes for u64, got {len(raw)}")
    return int.from_bytes(raw, "big", signed=False)


def _write_count(buffer: BytesIO, count: int, *, large: bool) -> None:
    buffer.write(count.to_bytes(8 if large else 4, "big", signed=False))


def _is_large_count(count: int) -> bool:
    return count > _U32_MAX


# ============================================================================
# payload builders / readers
# ============================================================================

def _iter_items(buffer: BytesIO, count: int) -> Iterator[Serialized[object]]:
    for _ in range(count):
        start = buffer.tell()
        item = Serialized.read_from(buffer, pos=start)
        buffer.seek(item.head.payload_end)
        yield item


def _iter_entry_pairs(
    buffer: BytesIO,
    count: int,
) -> Iterator[tuple[Serialized[object], Serialized[object]]]:
    for _ in range(count):
        key_start = buffer.tell()
        key = Serialized.read_from(buffer, pos=key_start)
        buffer.seek(key.head.payload_end)

        value_start = buffer.tell()
        value = Serialized.read_from(buffer, pos=value_start)
        buffer.seek(value.head.payload_end)

        yield key, value


def _build_collection_payload(
    items: Iterator[object],
    *,
    count: int,
    large: bool,
) -> BytesIO:
    payload = BytesIO()
    _write_count(payload, count, large=large)
    for item in items:
        Serialized.from_python_object(item).write_to(payload)
    return payload


def _build_mapping_payload(
    items: Iterator[tuple[object, object]],
    *,
    count: int,
    large: bool,
) -> BytesIO:
    payload = BytesIO()
    _write_count(payload, count, large=large)
    for key, value in items:
        Serialized.from_python_object(key).write_to(payload)
        Serialized.from_python_object(value).write_to(payload)
    return payload


def _materialize_iterable(obj: AbcIterator[object]) -> tuple[tuple[object, ...], int]:
    """
    Snapshot a one-shot iterator/generator into an immutable tuple so the wire
    format can include an upfront count.
    """
    values = tuple(obj)
    return values, len(values)


def _build_sequence_serialized(
    *,
    tag_small: int,
    tag_large: int,
    items: Iterator[object],
    count: int,
    metadata: Mapping[bytes, bytes] | None,
    codec: int | None,
) -> Serialized[object]:
    large = _is_large_count(count)
    payload = _build_collection_payload(items, count=count, large=large)
    return Serialized.build(
        tag=tag_large if large else tag_small,
        data=payload.to_bytes(),
        metadata=metadata,
        codec=codec,
    )


def _build_mapping_serialized(
    *,
    items: Iterator[tuple[object, object]],
    count: int,
    metadata: Mapping[bytes, bytes] | None,
    codec: int | None,
) -> Serialized[object]:
    large = _is_large_count(count)
    payload = _build_mapping_payload(items, count=count, large=large)
    return Serialized.build(
        tag=Tags.LARGE_MAPPING if large else Tags.MAPPING,
        data=payload.to_bytes(),
        metadata=metadata,
        codec=codec,
    )


# ============================================================================
# base classes
# ============================================================================

@dataclass(frozen=True, slots=True)
class CollectionSerialized(Serialized[T], Generic[T]):
    """
    Base class for count-prefixed collection payloads.

    Standard wire format
    --------------------
    Ordered collections:
        [count:u32][item_0][item_1]...[item_n]

    Mapping collections:
        [count:u32][key_0][value_0]...[key_n][value_n]
    """

    TAG: ClassVar[int]

    def _payload_buffer(self) -> BytesIO:
        if self.codec == CODEC_NONE:
            return BytesIO(self.data.to_bytes())
        return BytesIO(self.decode())

    def _read_count(self, buffer: BytesIO) -> int:
        return _read_u32(buffer)

    def _iter_from_payload(self) -> Iterator[Serialized[object]]:
        buf = self._payload_buffer()
        count = self._read_count(buf)
        yield from _iter_items(buf, count)

    def iter_(self) -> Iterator[Serialized[object]]:
        yield from self._iter_from_payload()

    @property
    def items(self) -> tuple[Serialized[object], ...]:
        return tuple(self.iter_())

    @property
    def value(self) -> T:
        raise NotImplementedError

    def as_python(self) -> T:
        return self.value

    @classmethod
    def from_python_object(
        cls,
        obj: object,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object] | None:
        # ------------------------------------------------------------------
        # concrete builtins / stdlib containers with known length
        # ------------------------------------------------------------------
        if isinstance(obj, list):
            return _build_sequence_serialized(
                tag_small=Tags.LIST,
                tag_large=Tags.LARGE_LIST,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, tuple):
            return _build_sequence_serialized(
                tag_small=Tags.TUPLE,
                tag_large=Tags.LARGE_TUPLE,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, set):
            return _build_sequence_serialized(
                tag_small=Tags.SET,
                tag_large=Tags.LARGE_SET,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, frozenset):
            return _build_sequence_serialized(
                tag_small=Tags.FROZENSET,
                tag_large=Tags.LARGE_FROZENSET,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, deque):
            return _build_sequence_serialized(
                tag_small=Tags.DEQUE,
                tag_large=Tags.LARGE_DEQUE,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, array):
            return _build_sequence_serialized(
                tag_small=Tags.ARRAY,
                tag_large=Tags.LARGE_ARRAY,
                items=iter(obj),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        # ------------------------------------------------------------------
        # mappings
        # ------------------------------------------------------------------
        if isinstance(obj, dict):
            return _build_mapping_serialized(
                items=iter(obj.items()),
                count=len(obj),
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, AbcMapping):
            items = tuple(obj.items())
            return _build_mapping_serialized(
                items=iter(items),
                count=len(items),
                metadata=metadata,
                codec=codec,
            )

        # ------------------------------------------------------------------
        # one-shot iterables
        # ------------------------------------------------------------------
        if isinstance(obj, AbcGenerator):
            values, count = _materialize_iterable(obj)
            return _build_sequence_serialized(
                tag_small=Tags.GENERATOR,
                tag_large=Tags.LARGE_GENERATOR,
                items=iter(values),
                count=count,
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, AbcIterator):
            values, count = _materialize_iterable(obj)
            return _build_sequence_serialized(
                tag_small=Tags.ITERATOR,
                tag_large=Tags.LARGE_ITERATOR,
                items=iter(values),
                count=count,
                metadata=metadata,
                codec=codec,
            )

        return None


@dataclass(frozen=True, slots=True)
class LargeCollectionSerialized(CollectionSerialized[T], Generic[T]):
    def _read_count(self, buffer: BytesIO) -> int:
        return _read_u64(buffer)


# ============================================================================
# sequence-like concrete serializers
# ============================================================================

@dataclass(frozen=True, slots=True)
class ArraySerialized(CollectionSerialized[list[object]]):
    TAG: ClassVar[int] = Tags.ARRAY

    @property
    def value(self) -> list[object]:
        return [item.as_python() for item in self.iter_()]


@dataclass(frozen=True, slots=True)
class LargeArraySerialized(LargeCollectionSerialized[list[object]]):
    TAG: ClassVar[int] = Tags.LARGE_ARRAY

    @property
    def value(self) -> list[object]:
        return [item.as_python() for item in self.iter_()]


@dataclass(frozen=True, slots=True)
class ListSerialized(CollectionSerialized[list[object]]):
    TAG: ClassVar[int] = Tags.LIST

    @property
    def value(self) -> list[object]:
        return [item.as_python() for item in self.iter_()]


@dataclass(frozen=True, slots=True)
class LargeListSerialized(LargeCollectionSerialized[list[object]]):
    TAG: ClassVar[int] = Tags.LARGE_LIST

    @property
    def value(self) -> list[object]:
        return [item.as_python() for item in self.iter_()]


@dataclass(frozen=True, slots=True)
class TupleSerialized(CollectionSerialized[tuple[object, ...]]):
    TAG: ClassVar[int] = Tags.TUPLE

    @property
    def value(self) -> tuple[object, ...]:
        return tuple(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class LargeTupleSerialized(LargeCollectionSerialized[tuple[object, ...]]):
    TAG: ClassVar[int] = Tags.LARGE_TUPLE

    @property
    def value(self) -> tuple[object, ...]:
        return tuple(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class SetSerialized(CollectionSerialized[set[object]]):
    TAG: ClassVar[int] = Tags.SET

    @property
    def value(self) -> set[object]:
        return {item.as_python() for item in self.iter_()}


@dataclass(frozen=True, slots=True)
class LargeSetSerialized(LargeCollectionSerialized[set[object]]):
    TAG: ClassVar[int] = Tags.LARGE_SET

    @property
    def value(self) -> set[object]:
        return {item.as_python() for item in self.iter_()}


@dataclass(frozen=True, slots=True)
class FrozenSetSerialized(CollectionSerialized[frozenset[object]]):
    TAG: ClassVar[int] = Tags.FROZENSET

    @property
    def value(self) -> frozenset[object]:
        return frozenset(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class LargeFrozenSetSerialized(LargeCollectionSerialized[frozenset[object]]):
    TAG: ClassVar[int] = Tags.LARGE_FROZENSET

    @property
    def value(self) -> frozenset[object]:
        return frozenset(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class DequeSerialized(CollectionSerialized[deque[object]]):
    TAG: ClassVar[int] = Tags.DEQUE

    @property
    def value(self) -> deque[object]:
        return deque(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class LargeDequeSerialized(LargeCollectionSerialized[deque[object]]):
    TAG: ClassVar[int] = Tags.LARGE_DEQUE

    @property
    def value(self) -> deque[object]:
        return deque(item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class GeneratorSerialized(CollectionSerialized[Iterator[object]]):
    TAG: ClassVar[int] = Tags.GENERATOR

    @property
    def value(self) -> Iterator[object]:
        return (item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class LargeGeneratorSerialized(LargeCollectionSerialized[Iterator[object]]):
    TAG: ClassVar[int] = Tags.LARGE_GENERATOR

    @property
    def value(self) -> Iterator[object]:
        return (item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class IteratorSerialized(CollectionSerialized[Iterator[object]]):
    TAG: ClassVar[int] = Tags.ITERATOR

    @property
    def value(self) -> Iterator[object]:
        return (item.as_python() for item in self.iter_())


@dataclass(frozen=True, slots=True)
class LargeIteratorSerialized(LargeCollectionSerialized[Iterator[object]]):
    TAG: ClassVar[int] = Tags.LARGE_ITERATOR

    @property
    def value(self) -> Iterator[object]:
        return (item.as_python() for item in self.iter_())


# ============================================================================
# mapping serializers
# ============================================================================

@dataclass(frozen=True, slots=True)
class _BaseMappingSerialized(CollectionSerialized[dict[object, object]]):
    def _iter_entry_pairs(self) -> Iterator[tuple[Serialized[object], Serialized[object]]]:
        buf = self._payload_buffer()
        count = self._read_count(buf)
        yield from _iter_entry_pairs(buf, count)

    def iter_(self) -> Iterator[Serialized[object]]:
        for key, value in self._iter_entry_pairs():
            yield key
            yield value

    def iter_entries(self) -> Iterator[tuple[Serialized[object], Serialized[object]]]:
        yield from self._iter_entry_pairs()

    @property
    def entries(self) -> tuple[tuple[Serialized[object], Serialized[object]], ...]:
        return tuple(self.iter_entries())

    @property
    def value(self) -> dict[object, object]:
        return {
            key.as_python(): value.as_python()
            for key, value in self.iter_entries()
        }


@dataclass(frozen=True, slots=True)
class MappingSerialized(_BaseMappingSerialized):
    TAG: ClassVar[int] = Tags.MAPPING


@dataclass(frozen=True, slots=True)
class LargeMappingSerialized(
    _BaseMappingSerialized,
    LargeCollectionSerialized[dict[object, object]],
):
    TAG: ClassVar[int] = Tags.LARGE_MAPPING


# ============================================================================
# registration
# ============================================================================

for cls in CollectionSerialized.__subclasses__():
    Tags.register_class(cls)

for cls in LargeCollectionSerialized.__subclasses__():
    Tags.register_class(cls)

for cls in _BaseMappingSerialized.__subclasses__():
    Tags.register_class(cls)

for t, cls in (
    (list, ListSerialized),
    (tuple, TupleSerialized),
    (set, SetSerialized),
    (frozenset, FrozenSetSerialized),
    (dict, MappingSerialized),
    (deque, DequeSerialized),
    (array, ArraySerialized),
):
    Tags.register_class(cls, pytype=t)


"""HTTP session abstraction with transparent Delta-table caching.

The :class:`Session` base class provides:

* A **thread-safe connection pool** backed by :class:`~yggdrasil.concurrent.threading.JobPoolExecutor`.
* **Single-request** dispatch via :meth:`~Session.send` (implemented by subclasses).
* **Concurrent batch dispatch** via :meth:`~Session.send_many`, with optional Delta
  table caching that de-duplicates live requests against previously stored responses.
* **Convenience HTTP-verb shortcuts**: :meth:`~Session.get`, :meth:`~Session.post`,
  :meth:`~Session.put`, :meth:`~Session.patch`, :meth:`~Session.delete`,
  :meth:`~Session.head`, :meth:`~Session.options`.
* **Spark-native scatter** via :meth:`~Session.spark_send` — distributes requests
  across Spark workers using a broadcast-session UDF.

Public method summary
---------------------

.. list-table::
   :header-rows: 1
   :widths: 25 75

   * - Method
     - Description
   * - :meth:`~Session.send`
     - Send one :class:`~yggdrasil.io.request.PreparedRequest`; must be
       implemented by subclasses.
   * - :meth:`~Session.send_many`
     - Send many requests in parallel, with optional Delta-table caching.
   * - :meth:`~Session.spark_send`
     - Distribute requests across a Spark cluster via a broadcast-session UDF.
   * - :meth:`~Session.request`
     - Build + dispatch a request from raw components (method, url, …).
   * - :meth:`~Session.prepare_request`
     - Resolve the URL and build a :class:`~yggdrasil.io.request.PreparedRequest`.
   * - ``get`` / ``post`` / ``put`` / ``patch`` / ``delete`` / ``head`` / ``options``
     - HTTP-verb shortcuts that delegate to :meth:`~Session.request`.

Configuration dataclasses
-------------------------
:class:`~yggdrasil.io.send_config.SendConfig`
    Controls a single :meth:`~Session.send` call (wait strategy, error handling,
    streaming, optional Delta cache).
:class:`~yggdrasil.io.send_config.SendManyConfig`
    Extends :class:`~yggdrasil.io.send_config.SendConfig` with batching and
    concurrency options for :meth:`~Session.send_many` and :meth:`~Session.spark_send`.

Config precedence rule
~~~~~~~~~~~~~~~~~~~~~~
When both a *config* object and explicit keyword arguments are supplied to any
send method, **explicit kwargs always win**.  This lets you share a base config
and override individual fields per-call without rebuilding the object::

    base = SendConfig(raise_error=False, wait=30)
    # raise_error is overridden to True for this one call:
    resp = session.get("/critical", config=base, raise_error=True)

Typical usage
-------------
::

    from yggdrasil.io.session import Session
    from yggdrasil.io.send_config import SendConfig, SendManyConfig

    # Single request with a context manager (ensures pool shutdown)
    with Session.from_url("https://api.example.com") as session:
        resp = session.get("/data", config=SendConfig(stream=False))

    # Batch with caching
    many_cfg = SendManyConfig(batch_size=50, ordered=True, cache=my_table)
    for resp in session.send_many(requests, config=many_cfg):
        process(resp)

    # Spark scatter
    spark_df = session.spark_send(requests, config=many_cfg, spark_session=spark)
"""
from __future__ import annotations

import base64
import datetime as dt
import itertools
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Union,
)

import pyarrow as pa
import yggdrasil.pickle.ser as pickle
from yggdrasil.concurrent.threading import Job, JobPoolExecutor
from yggdrasil.data import CastOptions, CastOptionsArg, any_to_datetime
from yggdrasil.dataclasses import restore_dataclass_state, serialize_dataclass_state
from yggdrasil.dataclasses.waiting import DEFAULT_WAITING_CONFIG, WaitingConfig, WaitingConfigArg
from yggdrasil.environ import PyEnv
from yggdrasil.io import SaveMode

from .buffer import BytesIO
from .request import PreparedRequest
from .response import RESPONSE_ARROW_SCHEMA, Response
from .send_config import SendConfig, SendManyConfig
from .url import URL

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

    from yggdrasil.databricks.sql.table import Table

__all__ = ["Session", "SendConfig", "SendManyConfig"]

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def to_utc_epoch_us(x: dt.datetime | dt.date | str) -> int:
    """Convert a date, datetime, or ISO-8601 string to UTC microseconds since epoch.

    Parameters
    ----------
    x:
        A :class:`datetime.date`, :class:`datetime.datetime`, or ISO-8601
        string.  Naive datetimes are assumed to be in UTC.

    Returns
    -------
    int
        Microseconds since the Unix epoch in UTC.
    """
    if isinstance(x, dt.date) and not isinstance(x, dt.datetime):
        v = dt.datetime(x.year, x.month, x.day, tzinfo=dt.timezone.utc)
    else:
        v = any_to_datetime(x)

    if v.tzinfo is None:
        v = v.replace(tzinfo=dt.timezone.utc)

    return int(v.astimezone(dt.timezone.utc).timestamp() * 1_000_000)


# ---------------------------------------------------------------------------
# Session
# ---------------------------------------------------------------------------

@dataclass
class Session(ABC):
    """Abstract base class for HTTP sessions.

    Subclasses must implement :meth:`send`.  Everything else — pooling,
    caching, batching, verb shortcuts, and Spark scatter — is provided here.

    Parameters
    ----------
    base_url:
        Optional base URL prepended to every relative path passed to
        :meth:`request` and the HTTP-verb shortcuts.  Parsed to a
        :class:`~yggdrasil.io.url.URL` on first use.
    verify:
        Whether to verify TLS certificates.  Passed through to the
        transport layer unchanged.
    pool_maxsize:
        Maximum number of concurrent worker threads in the job pool.
        Values ≤ 0 are silently coerced to 8.
    send_headers:
        Default headers merged into every outgoing request.  Useful for
        API keys, ``User-Agent``, or ``Content-Type`` applied globally.
        See also :attr:`x_api_key` for a typed ``X-API-Key`` shortcut.
    waiting:
        Default :class:`~yggdrasil.dataclasses.waiting.WaitingConfig`
        used when ``wait=None`` is passed to :meth:`send`.  Controls
        timeout, retry interval, back-off factor and maximum retries.

    Notes
    -----
    **Config precedence** — when both a *config* object and explicit keyword
    arguments are provided to any send method, the explicit kwargs always win.
    The config is only consulted for fields whose kwargs equal the method
    default (i.e. the caller did not override them).

    **Thread safety** — the connection pool and job pool are lazily
    initialised under a :class:`threading.RLock`, so the session is safe to
    share across threads.  The ``send_headers`` dict is *not* protected by a
    lock; if you mutate it from multiple threads, add your own synchronisation.

    **Pickle / Spark** — :meth:`__getstate__` / :meth:`__setstate__` are
    implemented via :func:`~yggdrasil.dataclasses.serialize_dataclass_state`,
    so the session can be pickled and broadcast to Spark executors.  The
    internal connection pool and lock are re-created on unpickling.

    Examples
    --------
    Open a session with a context manager (ensures pool shutdown on exit)::

        with Session.from_url("https://api.example.com") as s:
            resp = s.get("/v1/items")

    Share a base :class:`~yggdrasil.io.send_config.SendConfig` and override
    individual fields per-call::

        cfg = SendConfig(raise_error=False, stream=False)
        resp = s.get("/v1/items", config=cfg)
        if not resp.ok:
            log_error(resp)

    Batch dispatch with Delta-table caching::

        many_cfg = SendManyConfig(
            cache=my_delta_table,
            cache_by=["request_url_path", "request_body_hash"],
            batch_size=100,
            ordered=False,
        )
        for resp in s.send_many(iter(prepared_requests), config=many_cfg):
            handle(resp)
    """

    base_url: Optional[URL] = None
    verify: bool = True
    pool_maxsize: int = 10
    send_headers: Optional[dict[str, str]] = field(default=None, repr=False)
    waiting: WaitingConfig = field(
        default_factory=lambda: DEFAULT_WAITING_CONFIG,
        repr=False,
        compare=False,
        hash=False,
    )

    _lock: threading.RLock = field(default=None, init=False, repr=False, compare=False)
    _job_pool: JobPoolExecutor = field(default=None, init=False, repr=False, compare=False)

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    def __post_init__(self) -> None:
        """Normalise fields and initialise internal state after construction.

        * Parses :attr:`base_url` to a :class:`~yggdrasil.io.url.URL` if a
          plain string was supplied.
        * Creates the :class:`threading.RLock` used to guard lazy
          initialisation of the connection and job pools.
        * Coerces :attr:`pool_maxsize` ≤ 0 to 8.
        """
        if self.base_url:
            self.base_url = URL.parse(self.base_url)
        if self._lock is None:
            self._lock = threading.RLock()
        if self.pool_maxsize <= 0:
            self.pool_maxsize = 8

    def __getstate__(self) -> dict:
        """Return a picklable state dict (excludes lock and pools).

        Uses :func:`~yggdrasil.dataclasses.serialize_dataclass_state` so that
        non-picklable fields (``_lock``, ``_job_pool``) are excluded, enabling
        the session to be broadcast to Spark executors.
        """
        return serialize_dataclass_state(self)

    def __setstate__(self, state: dict) -> None:
        """Restore the session from a pickled state dict.

        Calls :meth:`__post_init__` after restoring fields so that the lock
        and pools are re-created fresh on the receiving process / executor.
        """
        restore_dataclass_state(self, state)
        self.__post_init__()

    def __enter__(self) -> "Session":
        """Support use as a context manager; returns *self*."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Shut down the job pool and release worker threads on context exit."""
        if self._job_pool:
            self._job_pool.shutdown(wait=True)
            self._job_pool = None

    # ------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------

    @property
    def job_pool(self) -> JobPoolExecutor:
        """Lazily-initialised thread pool (double-checked locking)."""
        if self._job_pool is None:
            with self._lock:
                if self._job_pool is None:
                    self._job_pool = JobPoolExecutor(max_workers=self.pool_maxsize)
        return self._job_pool

    @property
    def x_api_key(self) -> Optional[str]:
        """Read the ``X-API-Key`` header from :attr:`send_headers`, if present."""
        if self.send_headers:
            return self.send_headers.get("X-API-Key")
        return None

    @x_api_key.setter
    def x_api_key(self, value: Optional[str]) -> None:
        """Set the ``X-API-Key`` header in :attr:`send_headers`."""
        if value:
            if not self.send_headers:
                self.send_headers = {}
            self.send_headers["X-API-Key"] = value

    # ------------------------------------------------------------------
    # Construction
    # ------------------------------------------------------------------

    @classmethod
    def from_url(
        cls,
        url: Union[URL, str],
        *,
        verify: bool = True,
        normalize: bool = True,
        waiting: WaitingConfigArg = True,
    ) -> "Session":
        """Construct the appropriate :class:`Session` subclass for *url*.

        Currently supports ``http://`` and ``https://`` schemes, returning an
        :class:`~yggdrasil.io.http_.HTTPSession`.

        Parameters
        ----------
        url:
            Target base URL.
        verify:
            Whether to verify TLS certificates.
        normalize:
            Whether to normalize the URL (e.g. lower-case scheme/host).
        waiting:
            Default retry/wait configuration for the session.

        Returns
        -------
        Session
            A concrete session instance ready to send requests.

        Raises
        ------
        ValueError
            If the URL scheme is not supported.
        """
        url = URL.parse(url, normalize=normalize)

        if url.scheme.startswith("http"):
            from .http_ import HTTPSession

            return HTTPSession(
                base_url=url,
                verify=verify,
                waiting=WaitingConfig.check_arg(waiting) if waiting is not None else None,
            )

        raise ValueError(f"Cannot build session from scheme: {url.scheme!r}")

    # ------------------------------------------------------------------
    # Abstract transport
    # ------------------------------------------------------------------

    @abstractmethod
    def send(
        self,
        request: PreparedRequest,
        *,
        config: Optional[SendConfig] = None,
        wait: WaitingConfigArg = None,
        raise_error: bool = True,
        stream: bool = True,
        cache: Optional["Table"] = None,
        cache_by: Optional[list[str]] = None,
        received_from: Optional[dt.datetime | dt.date | str] = None,
        received_to: Optional[dt.datetime | dt.date | str] = None,
        anonymize: Literal["remove", "redact"] = "remove",
        wait_cache: WaitingConfigArg = False,
    ) -> Response:
        """Send a single prepared request and return its response.

        This is the only method subclasses **must** implement.

        Parameters
        ----------
        request:
            The fully prepared request to dispatch.
        config:
            Optional :class:`SendConfig` providing defaults for every other
            keyword argument.  Explicit kwargs override the config.
        wait:
            Retry / waiting strategy.  ``None`` defers to ``config.wait``
            or the session's :attr:`waiting` default.
        raise_error:
            Raise on non-2xx responses when ``True`` (default).
        stream:
            Stream the response body lazily when ``True`` (default).
        cache:
            Delta table used to cache responses.  A cache hit returns the
            stored response without hitting the network.
        cache_by:
            Column names that form the cache lookup key.  Defaults to the
            standard request-fingerprint columns when ``cache`` is set.
        received_from:
            Earliest acceptable cached response timestamp.
        received_to:
            Latest acceptable cached response timestamp.
        anonymize:
            How to strip sensitive fields before the cache lookup
            (``"remove"`` or ``"redact"``).
        wait_cache:
            Waiting config for the background cache-write.  ``False``
            means fire-and-forget.

        Returns
        -------
        Response
            The HTTP response.

        Raises
        ------
        NotImplementedError
            Always — subclasses must override this method.
        """
        raise NotImplementedError

    # ------------------------------------------------------------------
    # Cache helpers (private)
    # ------------------------------------------------------------------

    @staticmethod
    def _cache_by_keys(arg: Optional[list[str]] = None) -> list[str]:
        """Resolve and validate cache-key column names.

        Parameters
        ----------
        arg:
            User-supplied column list.  ``None`` or empty falls back to the
            standard request fingerprint columns.

        Returns
        -------
        list[str]
            Validated list of column names present in
            :data:`~yggdrasil.io.response.RESPONSE_ARROW_SCHEMA`.

        Raises
        ------
        ValueError
            If any name is not in the response Arrow schema.
        """
        if not arg:
            arg = [
                "request_method",
                "request_url_host",
                "request_url_path",
                "request_url_query",
                "request_content_length",
                "request_body_hash",
            ]

        invalid = [k for k in arg if k not in RESPONSE_ARROW_SCHEMA.names]
        if invalid:
            raise ValueError(
                f"Invalid cache_by key(s): {invalid!r}. "
                f"Must be within: {RESPONSE_ARROW_SCHEMA.names}"
            )
        return arg

    @staticmethod
    def _cache_value_from_request(request: PreparedRequest, key: str) -> Any:
        """Extract the value of a single cache-key field from a prepared request.

        A fixed dispatch table handles the standard request-fingerprint fields.
        Any field not in the table falls back to ``getattr(request, key)``,
        allowing future :class:`~yggdrasil.io.request.PreparedRequest`
        extensions without changing this method.

        Parameters
        ----------
        request:
            The prepared request to inspect.
        key:
            One of the recognised cache-key field names (e.g.
            ``"request_method"``, ``"request_url_host"``,
            ``"request_body_hash"``).

        Returns
        -------
        Any
            The extracted value, or ``None`` if the field is absent / empty.

        Raises
        ------
        ValueError
            If *key* is not in the dispatch table and not an attribute of
            *request*.
        """
        _EXTRACTORS: dict[str, Callable[[PreparedRequest], Any]] = {
            "request_method":         lambda r: r.method,
            "request_url":            lambda r: r.url.to_string(),
            "request_url_scheme":     lambda r: r.url.scheme,
            "request_url_host":       lambda r: r.url.host,
            "request_url_port":       lambda r: r.url.port,
            "request_url_path":       lambda r: r.url.path,
            "request_url_query":      lambda r: r.url.query,
            "request_body_hash":      lambda r: r.body.xxh3_int64() if r.body else None,
            "request_content_length": lambda r: r.content_length,
        }
        extractor = _EXTRACTORS.get(key)
        if extractor is not None:
            return extractor(request)
        # Generic fallback for future PreparedRequest attributes
        if hasattr(request, key):
            return getattr(request, key)
        raise ValueError(f"Unsupported request cache_by key: {key!r}")

    @classmethod
    def _cache_values_from_request(
        cls,
        request: PreparedRequest,
        keys: list[str],
    ) -> dict[str, Any]:
        """Extract multiple cache-key values from *request* into a mapping.

        Parameters
        ----------
        request:
            The prepared request to inspect.
        keys:
            Ordered list of cache-key field names.

        Returns
        -------
        dict[str, Any]
            ``{key: value}`` for every key in *keys*, preserving order.
        """
        return {k: cls._cache_value_from_request(request, k) for k in keys}

    @classmethod
    def _cache_value_from_response(cls, response: Response, key: str) -> Any:
        """Extract a single cache-key value from a *response* or its embedded request.

        Parameters
        ----------
        response:
            The response to inspect.  If it carries an embedded
            :class:`~yggdrasil.io.request.PreparedRequest`, request-prefixed
            keys (``"request_*"``) are resolved from that object.
        key:
            Cache-key field name.

        Returns
        -------
        Any

        Raises
        ------
        ValueError
            If *key* cannot be resolved from the response or its request.
        """
        if hasattr(response, key):
            return getattr(response, key)
        if key.startswith("request_"):
            return cls._cache_value_from_request(response.request, key)
        raise ValueError(f"Unsupported response cache_by key: {key!r}")

    @classmethod
    def _cache_tuple_from_request(
        cls,
        request: PreparedRequest,
        keys: list[str],
    ) -> tuple:
        """Build a hashable cache-key tuple from a request.

        Parameters
        ----------
        request:
            The prepared request.
        keys:
            Ordered list of cache-key field names.  The tuple preserves
            this order so it can be compared against tuples built from
            :meth:`_cache_tuple_from_response` using the same *keys*.

        Returns
        -------
        tuple
            ``(value_for_keys[0], value_for_keys[1], …)``
        """
        values = cls._cache_values_from_request(request, keys)
        return tuple(values[k] for k in keys)

    @classmethod
    def _cache_tuple_from_response(
        cls,
        response: Response,
        keys: list[str],
    ) -> tuple:
        """Build a hashable cache-key tuple from a response.

        Parameters
        ----------
        response:
            The response (may embed a request for ``"request_*"`` keys).
        keys:
            Ordered list of cache-key field names.

        Returns
        -------
        tuple
            ``(value_for_keys[0], value_for_keys[1], …)``
        """
        return tuple(cls._cache_value_from_response(response, k) for k in keys)

    @staticmethod
    def _sql_literal(value: Any) -> str:
        """Format *value* as a SQL literal safe for embedding in a WHERE clause.

        Handles the types that appear in the response cache schema:

        * ``None`` → ``null``
        * ``bool`` → ``true`` / ``false``  (checked before ``int`` because
          :class:`bool` is a subclass of :class:`int`)
        * ``int`` / ``float`` → bare numeric string
        * :class:`datetime.datetime` → ``timestamp 'YYYY-MM-DD HH:MM:SS.ffffff'``
        * ``bytes`` → base-64 encoded, single-quoted string
        * anything else → ``str(value)``, single-quoted with ``'`` escaped as ``''``

        Parameters
        ----------
        value:
            The Python value to serialise.

        Returns
        -------
        str
            A SQL literal fragment (no trailing semicolon).
        """
        if value is None:
            return "null"
        if isinstance(value, bool):
            return "true" if value else "false"
        if isinstance(value, (int, float)):
            return str(value)
        if isinstance(value, dt.datetime):
            return f"timestamp '{value.isoformat(sep=' ', timespec='microseconds')}'"
        if isinstance(value, bytes):
            value = base64.b64encode(value).decode("ascii")
        else:
            value = str(value)
        return f"'{value.replace(chr(39), chr(39) * 2)}'"

    @classmethod
    def _sql_match_clause(
        cls,
        request: Optional[PreparedRequest],
        keys: list[str],
        received_from: Optional[dt.datetime | dt.date | str] = None,
        received_to: Optional[dt.datetime | dt.date | str] = None,
    ) -> str:
        """Build a SQL WHERE clause that matches *request* against the cache table.

        Parameters
        ----------
        request:
            The request whose field values become equality predicates.
            Pass ``None`` to skip request predicates (time-range only).
        keys:
            Cache-key column names to include in the WHERE clause.
        received_from:
            Lower bound on ``response_received_at_epoch`` (inclusive).
        received_to:
            Upper bound on ``response_received_at_epoch`` (inclusive).

        Returns
        -------
        str
            A SQL fragment suitable for use after ``WHERE``.  Returns
            ``"1=1"`` when no predicates are required.
        """
        clauses: list[str] = []

        if request is not None and keys:
            for key, value in cls._cache_values_from_request(request, keys).items():
                if value is None:
                    clauses.append(f"{key} IS NULL")
                else:
                    clauses.append(f"{key} = {cls._sql_literal(value)}")

        if received_from not in (None, ""):
            clauses.append(
                f"response_received_at_epoch >= {to_utc_epoch_us(received_from)}"
            )
        if received_to not in (None, ""):
            clauses.append(
                f"response_received_at_epoch <= {to_utc_epoch_us(received_to)}"
            )

        return " AND ".join(clauses) if clauses else "1=1"

    # ------------------------------------------------------------------
    # Config resolution helpers (private)
    # ------------------------------------------------------------------

    @staticmethod
    def _resolve_send_config(
        config: Optional[SendConfig],
        *,
        wait: WaitingConfigArg,
        raise_error: bool,
        stream: bool,
        cache: Optional["Table"],
        cache_by: Optional[list[str]],
        anonymize: Literal["remove", "redact"],
        received_from: Optional[dt.datetime | dt.date | str],
        received_to: Optional[dt.datetime | dt.date | str],
        wait_cache: WaitingConfigArg,
    ) -> SendConfig:
        """Merge *config* with explicit keyword arguments into a resolved :class:`SendConfig`.

        Explicit kwargs always win over the config object.  A ``None`` config
        is equivalent to :meth:`SendConfig.default`.

        Parameters
        ----------
        config:
            Base config or ``None``.
        wait, raise_error, stream, cache, cache_by, anonymize,
        received_from, received_to, wait_cache:
            Call-site keyword arguments that take precedence over the config.

        Returns
        -------
        SendConfig
            Fully resolved configuration with no ``None`` ambiguity.
        """
        base = config or SendConfig.default()
        return SendConfig(
            wait=wait if wait is not None else base.wait,
            raise_error=raise_error if raise_error is not True else base.raise_error,
            stream=stream if stream is not True else base.stream,
            cache=cache if cache is not None else base.cache,
            cache_by=cache_by if cache_by is not None else base.cache_by,
            anonymize=anonymize if anonymize != "remove" else base.anonymize,
            received_from=received_from if received_from is not None else base.received_from,
            received_to=received_to if received_to is not None else base.received_to,
            wait_cache=wait_cache if wait_cache is not False else base.wait_cache,
        )

    @staticmethod
    def _resolve_send_many_config(
        config: Optional[SendManyConfig],
        *,
        wait: WaitingConfigArg,
        raise_error: bool,
        normalize: Optional[bool],
        stream: bool,
        cache: Optional["Table"],
        cache_by: Optional[list[str]],
        cache_anonymize: Literal["remove", "redact"],
        received_from: Optional[dt.datetime | dt.date | str],
        received_to: Optional[dt.datetime | dt.date | str],
        wait_cache: WaitingConfigArg,
        batch_size: Optional[int],
        ordered: bool,
        max_in_flight: Optional[int],
    ) -> SendManyConfig:
        """Merge *config* with explicit keyword arguments into a resolved :class:`SendManyConfig`.

        Explicit kwargs always win over the config object.  A ``None`` config
        is equivalent to :meth:`SendManyConfig.default`.

        Parameters
        ----------
        config:
            Base config or ``None``.
        wait, raise_error, normalize, stream, cache, cache_by, cache_anonymize,
        received_from, received_to, wait_cache, batch_size, ordered, max_in_flight:
            Call-site keyword arguments that take precedence over the config.

        Returns
        -------
        SendManyConfig
            Fully resolved configuration with no ``None`` ambiguity.
        """
        base = config or SendManyConfig.default()
        return SendManyConfig(
            wait=wait if wait is not None else base.wait,
            raise_error=raise_error if raise_error is not True else base.raise_error,
            normalize=normalize if normalize is not None else base.normalize,
            stream=stream if stream is not True else base.stream,
            cache=cache if cache is not None else base.cache,
            cache_by=cache_by if cache_by is not None else base.cache_by,
            cache_anonymize=cache_anonymize if cache_anonymize != "remove" else base.cache_anonymize,
            received_from=received_from if received_from is not None else base.received_from,
            received_to=received_to if received_to is not None else base.received_to,
            wait_cache=wait_cache if wait_cache is not False else base.wait_cache,
            batch_size=batch_size if batch_size is not None else base.batch_size,
            ordered=ordered if ordered is not False else base.ordered,
            max_in_flight=max_in_flight if max_in_flight is not None else base.max_in_flight,
        )

    # ------------------------------------------------------------------
    # Batch dispatch with optional Delta-table caching
    # ------------------------------------------------------------------

    def send_many(
        self,
        requests: Iterator[PreparedRequest],
        *,
        config: Optional[SendManyConfig] = None,
        wait: WaitingConfigArg = None,
        raise_error: bool = True,
        normalize: Optional[bool] = None,
        stream: bool = True,
        cache: Optional["Table"] = None,
        cache_by: Optional[list[str]] = None,
        cache_anonymize: Literal["remove", "redact"] = "remove",
        received_from: Optional[dt.datetime | dt.date | str] = None,
        received_to: Optional[dt.datetime | dt.date | str] = None,
        wait_cache: WaitingConfigArg = False,
        batch_size: Optional[int] = None,
        ordered: bool = False,
        max_in_flight: Optional[int] = None,
    ) -> Iterator[Response]:
        """Send multiple requests concurrently, with optional Delta-table caching.

        Requests are dispatched in parallel up to :attr:`pool_maxsize`.  When
        a *cache* table is supplied, each batch is first looked up in the
        table; only cache misses hit the network.  Successful live responses
        are written back to the cache asynchronously.

        Parameters
        ----------
        requests:
            Iterator of :class:`PreparedRequest` objects to dispatch.
        config:
            Optional :class:`SendManyConfig` providing defaults for every
            other keyword argument.  Explicit kwargs override the config.
        wait:
            Retry / waiting strategy forwarded to each :meth:`send` call.
            ``None`` defers to ``config.wait`` or the session default.
        raise_error:
            When ``True`` (default), any failed response raises after all
            in-flight requests in the current batch complete.
        normalize:
            Whether to normalise request URLs.  ``None`` (default) enables
            normalisation automatically when ``cache`` is set.
        stream:
            Whether to stream response bodies lazily.
        cache:
            Delta table used as a response cache.  Cache hits are returned
            immediately; misses are fetched and written back.
        cache_by:
            Column names forming the cache key.  Defaults to the standard
            request fingerprint when ``cache`` is set.
        cache_anonymize:
            How to strip sensitive fields before cache reads/writes
            (``"remove"`` or ``"redact"``).
        received_from:
            Only return cached responses received on or after this timestamp.
        received_to:
            Only return cached responses received on or before this timestamp.
        wait_cache:
            Waiting config for background cache writes.  ``False`` is
            fire-and-forget.
        batch_size:
            Requests per cache-lookup batch.  ``None`` defaults to
            ``pool_maxsize × 100``.
        ordered:
            When ``True``, yield responses in input order.  ``False``
            (default) yields in completion order for higher throughput.
        max_in_flight:
            Maximum concurrent network requests.  ``None`` lets the pool
            decide.

        Yields
        ------
        Response
            Successful responses (or all responses when ``raise_error=False``).

        Raises
        ------
        Exception
            Re-raises the last failure in a batch when ``raise_error=True``
            and at least one request fails.
        """
        # Merge config + explicit kwargs → single resolved config
        cfg = self._resolve_send_many_config(
            config,
            wait=wait,
            raise_error=raise_error,
            normalize=normalize,
            stream=stream,
            cache=cache,
            cache_by=cache_by,
            cache_anonymize=cache_anonymize,
            received_from=received_from,
            received_to=received_to,
            wait_cache=wait_cache,
            batch_size=batch_size,
            ordered=ordered,
            max_in_flight=max_in_flight,
        )
        wait          = cfg.wait
        raise_error   = cfg.raise_error
        normalize     = cfg.normalize
        stream        = cfg.stream
        cache         = cfg.cache
        cache_by      = cfg.cache_by
        cache_anonymize = cfg.cache_anonymize
        received_from = cfg.received_from
        received_to   = cfg.received_to
        wait_cache    = cfg.wait_cache
        batch_size    = cfg.batch_size
        ordered       = cfg.ordered
        max_in_flight = cfg.max_in_flight

        if normalize is None:
            normalize = cache is not None

        # Resolve & validate cache keys up front
        if cache is not None:
            cache_by = self._cache_by_keys(cache_by)
            cache_request_by = [k for k in cache_by if k.startswith("request")]
        else:
            cache_request_by = []

        pool = self.job_pool
        if not batch_size:
            batch_size = pool.max_workers * 100

        # ----------------------------------------------------------
        # Fast path: no caching — just fan out all requests
        # ----------------------------------------------------------
        if cache is None:
            def _uncached_jobs() -> Iterator[Job]:
                for req in requests:
                    yield Job.make(
                        self.send,
                        req,
                        wait=wait,
                        raise_error=raise_error,
                        stream=stream,
                    )

            for result in pool.as_completed(
                _uncached_jobs(),
                ordered=ordered,
                max_in_flight=self.pool_maxsize,
                cancel_on_exit=True,
                shutdown_on_exit=True,
                raise_error=True,
            ):
                resp: Response = result.result
                if raise_error:
                    resp.raise_for_status()
                yield resp
            return

        # ----------------------------------------------------------
        # Cached path: batch → lookup → miss → fetch → write back
        # ----------------------------------------------------------
        def _batched(it: Iterator, n: int) -> Iterator[list]:
            it = iter(it)
            while True:
                batch = list(itertools.islice(it, n))
                if not batch:
                    break
                yield batch

        for batch in _batched(requests, batch_size):
            # Anonymise the batch for the cache lookup
            anon_batch = [
                req.anonymize(mode=cache_anonymize) if cache_anonymize else req
                for req in batch
            ]

            # Build a compound WHERE clause: (req1 OR req2 OR …) AND time_range
            request_clauses = " OR ".join(
                f"({self._sql_match_clause(req, keys=cache_request_by)})"
                for req in anon_batch
            )
            time_clause = self._sql_match_clause(
                None,
                keys=[],
                received_from=received_from,
                received_to=received_to,
            )
            where_parts: list[str] = []
            if request_clauses:
                where_parts.append(f"({request_clauses})")
            if time_clause and time_clause != "1=1":
                where_parts.append(f"({time_clause})")

            base_query = f"SELECT * FROM {cache.full_name(safe=True)}"
            if where_parts:
                base_query += " WHERE " + " AND ".join(where_parts)

            # Deduplicate: keep only the most-recent response per cache key
            if cache_request_by:
                partition_by = ", ".join(cache_request_by)
                query = (
                    f"SELECT * FROM ("
                    f"  SELECT t.*, row_number() OVER ("
                    f"    PARTITION BY {partition_by}"
                    f"    ORDER BY response_received_at_epoch DESC"
                    f"  ) AS __rn"
                    f"  FROM ({base_query}) t"
                    f") ranked WHERE __rn = 1"
                )
            else:
                query = base_query

            # Query cache (auto-create table on first run if missing)
            try:
                cache_result = cache.sql.execute(query)
            except Exception as exc:
                if "TABLE_OR_VIEW_NOT_FOUND" in str(exc):
                    cache.create(RESPONSE_ARROW_SCHEMA, if_not_exists=True)
                    cache_result = cache.sql.execute(query)
                else:
                    raise

            # Parse cached responses into a lookup map
            cached_arrow = cache_result.to_arrow_table()
            cached_responses = list(Response.from_arrow(cached_arrow))
            cache_map: dict[tuple, Response] = {
                self._cache_tuple_from_response(r, cache_request_by): r
                for r in cached_responses
            }

            # Partition the batch into hits and misses
            hits: list[Response] = []
            misses: list[PreparedRequest] = []
            for req in batch:
                anon_req = req.anonymize(mode="remove")
                key = self._cache_tuple_from_request(anon_req, cache_request_by)
                if key in cache_map:
                    hits.append(cache_map[key])
                else:
                    misses.append(req)

            # Yield cache hits immediately
            yield from hits

            if not misses:
                continue

            # Fetch misses from the network
            def _miss_jobs() -> Iterator[Job]:
                for req in misses:
                    yield Job.make(
                        self.send,
                        req,
                        wait=wait,
                        raise_error=False,
                        stream=stream,
                        cache=None,
                    )

            to_insert: list[Response] = []
            failed: list[Response] = []

            for result in pool.as_completed(
                _miss_jobs(),
                ordered=ordered,
                max_in_flight=max_in_flight,
                cancel_on_exit=False,
                shutdown_on_exit=False,
                raise_error=True,
            ):
                resp = result.result
                if resp.ok:
                    to_insert.append(resp)
                    yield resp
                elif raise_error:
                    failed.append(resp)

            # Write successful responses back to the cache
            if to_insert:
                batches = [
                    r.anonymize(mode="remove").to_arrow_batch(parse=False)
                    for r in to_insert
                ]
                combined = pa.Table.from_batches(batches).combine_chunks()
                cache.insert(
                    combined,
                    mode=SaveMode.APPEND,
                    match_by=cache_by,
                    wait=wait_cache,
                )

            if raise_error and failed:
                failed[-1].raise_for_status()

    # ------------------------------------------------------------------
    # Convenience HTTP-verb methods
    # ------------------------------------------------------------------

    def get(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a GET request.

        Parameters
        ----------
        url:
            Relative or absolute URL.  Relative URLs are resolved against
            :attr:`base_url`.
        config:
            Optional :class:`SendConfig` for this request.
        params:
            Query-string parameters merged into *url*.
        headers:
            Additional headers merged with :attr:`send_headers`.
        body:
            Optional request body.
        tags:
            Arbitrary string tags attached to the prepared request.
        stream:
            Stream the response body lazily (default ``True``).
        wait:
            Retry / waiting strategy.
        normalize:
            Normalise the URL before sending.
        cache:
            Delta table for response caching.
        **kwargs:
            Extra keyword arguments forwarded to :meth:`request`.
        """
        return self.request(
            "GET", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, stream=stream, wait=wait, normalize=normalize,
            cache=cache, **kwargs,
        )

    def post(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a POST request.  See :meth:`get` for parameter documentation."""
        return self.request(
            "POST", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, json=json, stream=stream, wait=wait,
            normalize=normalize, cache=cache, **kwargs,
        )

    def put(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a PUT request.  See :meth:`get` for parameter documentation."""
        return self.request(
            "PUT", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, json=json, stream=stream, wait=wait,
            normalize=normalize, cache=cache, **kwargs,
        )

    def patch(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a PATCH request.  See :meth:`get` for parameter documentation."""
        return self.request(
            "PATCH", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, json=json, stream=stream, wait=wait,
            normalize=normalize, cache=cache, **kwargs,
        )

    def delete(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a DELETE request.  See :meth:`get` for parameter documentation."""
        return self.request(
            "DELETE", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, json=json, stream=stream, wait=wait,
            normalize=normalize, cache=cache, **kwargs,
        )

    def head(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        stream: bool = False,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a HEAD request (``stream=False`` by default).

        See :meth:`get` for parameter documentation.
        """
        return self.request(
            "HEAD", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, stream=stream, wait=wait, normalize=normalize,
            cache=cache, **kwargs,
        )

    def options(
        self,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
        **kwargs: Any,
    ) -> Response:
        """Send an OPTIONS request.  See :meth:`get` for parameter documentation."""
        return self.request(
            "OPTIONS", url,
            config=config, params=params, headers=headers, body=body,
            tags=tags, json=json, stream=stream, wait=wait,
            normalize=normalize, cache=cache, **kwargs,
        )

    # ------------------------------------------------------------------
    # Request orchestration
    # ------------------------------------------------------------------

    def request(
        self,
        method: str,
        url: Optional[Union[URL, str]] = None,
        *,
        config: Optional[SendConfig] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        before_send: Optional[Callable[[PreparedRequest], PreparedRequest]] = None,
        json: Optional[Any] = None,
        stream: bool = True,
        wait: WaitingConfigArg = None,
        normalize: bool = True,
        cache: Optional["Table"] = None,
    ) -> Response:
        """Build and dispatch an HTTP request.

        Combines URL resolution, header merging, body serialisation, and
        :meth:`send` dispatch into a single call.  The HTTP-verb shortcuts
        (:meth:`get`, :meth:`post`, …) delegate here.

        Parameters
        ----------
        method:
            HTTP method string (e.g. ``"GET"``, ``"POST"``).
        url:
            Relative or absolute URL.
        config:
            Optional :class:`SendConfig` providing defaults.
        params:
            Query-string parameters merged into *url*.
        headers:
            Additional request headers.
        body:
            Raw request body.
        tags:
            Arbitrary string tags attached to the prepared request.
        before_send:
            Optional hook called on the :class:`PreparedRequest` immediately
            before dispatch.  Useful for signing or adding per-request headers.
        json:
            JSON-serialisable body.  Mutually exclusive with *body*.
        stream:
            Stream the response body lazily.
        wait:
            Retry / waiting strategy.
        normalize:
            Normalise the URL.  ``None`` enables normalisation when
            *cache* is set.
        cache:
            Delta table for response caching.

        Returns
        -------
        Response
        """
        if normalize is None:
            normalize = cache is not None

        prepared = self.prepare_request(
            method=method,
            url=url,
            params=params,
            headers=headers,
            body=body,
            tags=tags,
            json=json,
            normalize=normalize,
            before_send=before_send,
        )
        return self.send(
            prepared,
            config=config,
            stream=stream,
            wait=wait,
            cache=cache,
        )

    def prepare_request(
        self,
        method: str,
        url: Optional[Union[URL, str]] = None,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        body: Optional[Union[BytesIO, bytes]] = None,
        tags: Optional[Mapping[str, str]] = None,
        before_send: Optional[Callable[[PreparedRequest], PreparedRequest]] = None,
        after_received: Optional[Callable[[Response], Response]] = None,
        *,
        json: Optional[Any] = None,
        normalize: bool = True,
    ) -> PreparedRequest:
        """Resolve the URL and construct a :class:`PreparedRequest`.

        Parameters
        ----------
        method:
            HTTP method.
        url:
            Relative or absolute target.  Resolved against :attr:`base_url`
            when set.
        params:
            Query-string parameters merged into *url*.
        headers:
            Additional headers.
        body:
            Raw request body.
        tags:
            Arbitrary metadata attached to the request.
        before_send:
            Hook applied after preparation, before dispatch.
        after_received:
            Hook applied to the raw :class:`Response` after receipt.
        json:
            JSON body (mutually exclusive with *body*).
        normalize:
            Normalise the URL (default ``True``).

        Returns
        -------
        PreparedRequest

        Raises
        ------
        ValueError
            If both *url* and :attr:`base_url` are ``None``.
        """
        full_url: Union[URL, str, None] = url
        if self.base_url:
            full_url = self.base_url.join(url) if url else self.base_url
        elif url is None:
            raise ValueError("url is required when base_url is not set on the session.")

        if params:
            u = URL.parse(full_url, normalize=normalize)
            items = list(u.query_items(keep_blank_values=True))
            items.extend((k, v) for k, v in params.items())
            full_url = u.with_query_items(tuple(items))

        return PreparedRequest.prepare(
            method=method,
            url=full_url,
            headers=headers,
            body=body,
            tags=tags,
            json=json,
            normalize=normalize,
            before_send=before_send,
            after_received=after_received,
        )

    # ------------------------------------------------------------------
    # Spark scatter
    # ------------------------------------------------------------------

    def spark_send(
        self,
        requests: Iterator[PreparedRequest],
        *,
        parse: Union[CastOptionsArg, bool] = None,
        apply: Optional[Callable[[Response], Any]] = None,
        config: Optional[SendManyConfig] = None,
        spark_session: Optional["SparkSession"] = None,
        wait: WaitingConfigArg = None,
        raise_error: bool = True,
        normalize: Optional[bool] = None,
        stream: bool = True,
        cache: Optional["Table"] = None,
        cache_by: Optional[list[str]] = None,
        cache_anonymize: Literal["remove", "redact"] = "remove",
        received_from: Optional[dt.datetime | dt.date | str] = None,
        received_to: Optional[dt.datetime | dt.date | str] = None,
        wait_cache: WaitingConfigArg = False,
        batch_size: Optional[int] = None,
        ordered: bool = False,
        max_in_flight: Optional[int] = None,
    ):
        """Distribute requests across Spark workers via ``mapInArrow``.

        Requests are serialised as native Arrow rows conforming to
        :data:`~yggdrasil.io.request.REQUEST_ARROW_SCHEMA` — no pickling is
        involved on the input side.  Each executor deserialises its partition
        back into :class:`PreparedRequest` objects via
        :meth:`PreparedRequest.from_arrow`, sends them through
        :meth:`send_many`, and yields Arrow batches for the responses.

        The output schema is controlled by *parse*:

        ``parse=None`` / ``parse=False``
            Each response is serialised with
            ``resp.to_arrow_batch(parse=False)`` yielding the full
            :data:`~yggdrasil.io.response.RESPONSE_ARROW_SCHEMA` (request
            metadata + response metadata + raw body bytes).

        ``parse=True``
            The **first** request is sent eagerly on the driver to discover
            the parsed body Arrow schema.  Executors yield
            ``resp.to_arrow_batch(parse=True)`` — only the parsed body
            columns.

        ``parse=<CastOptionsArg>``
            A :class:`pa.Schema`, :class:`~yggdrasil.data.CastOptions`,
            etc.  The schema is extracted without a probe.  If it resolves
            to ``None``, falls back to ``RESPONSE_ARROW_SCHEMA`` (like
            ``parse=None``).

        Parameters
        ----------
        requests:
            Iterator of :class:`~yggdrasil.io.request.PreparedRequest`
            objects.  Consumed eagerly to build the input Arrow DataFrame.
        parse:
            Controls how responses are returned:

            * ``None`` / ``False`` — full response schema
              (:data:`RESPONSE_ARROW_SCHEMA`).
            * ``True`` — parsed body columns; schema discovered from a
              probe request.
            * Any :data:`~yggdrasil.data.CastOptionsArg` value — parsed
              body with an explicit schema.  Falls back to
              ``RESPONSE_ARROW_SCHEMA`` when no schema can be derived.
        apply:
            Optional transform applied to each :class:`Response` on the
            executor **before** Arrow serialisation.
        config:
            Optional :class:`~yggdrasil.io.send_config.SendManyConfig`.
            Explicit kwargs override the config.
        spark_session:
            Active :class:`pyspark.sql.SparkSession`.  When ``None`` one
            is created (or retrieved) automatically.
        wait:
            Retry / waiting strategy for each executor.
        raise_error:
            When ``True`` a non-2xx response raises inside the executor.
        normalize:
            Whether to normalise request URLs.
        stream:
            Whether to stream response bodies lazily on executors.
        cache:
            Delta table used as a response cache on each executor.
        cache_by:
            Column names forming the cache key.
        cache_anonymize:
            How sensitive fields are stripped before cache reads/writes.
        received_from:
            Earliest acceptable cached-response timestamp.
        received_to:
            Latest acceptable cached-response timestamp.
        wait_cache:
            Waiting config for background cache writes on executors.
        batch_size:
            Requests per cache-lookup batch on each executor.
        ordered:
            When ``True`` responses are returned in input order.
        max_in_flight:
            Maximum concurrent network requests per executor.

        Returns
        -------
        pyspark.sql.DataFrame
            When ``parse`` is ``None`` / ``False`` (or schema falls back):
            columns from :data:`RESPONSE_ARROW_SCHEMA`.

            When ``parse`` resolves to a body schema: columns from the
            parsed response body.

        Notes
        -----
        Because the session is broadcast, **all executors share the same
        session state**.  Mutable fields are copied at broadcast time.

        Examples
        --------
        Full response metadata (default)::

            df = session.spark_send(reqs)

        Parsed body — auto-discover schema::

            df = session.spark_send(reqs, parse=True)

        Parsed body — explicit schema::

            schema = pa.schema([("id", pa.int64()), ("name", pa.utf8())])
            df = session.spark_send(reqs, parse=schema)

        With a per-response transform::

            df = session.spark_send(reqs, parse=True, apply=my_transform)
        """
        from yggdrasil.spark.cast import arrow_schema_to_spark_schema
        from .request import REQUEST_ARROW_SCHEMA

        serialized_session = pickle.dumps(self)

        if spark_session is None:
            spark_session = PyEnv.spark_session(
                create=True, import_error=True, install_spark=True,
            )

        # --- Materialise requests into Arrow ----------------------------
        request_list = list(requests)
        request_batches = [req.to_arrow_batch() for req in request_list]

        if request_batches:
            request_table = pa.Table.from_batches(request_batches)
        else:
            request_table = pa.table(
                {f.name: pa.array([], type=f.type) for f in REQUEST_ARROW_SCHEMA},
            )

        request_spark_schema = arrow_schema_to_spark_schema(REQUEST_ARROW_SCHEMA)
        request_df = spark_session.createDataFrame(
            request_table.to_pandas(),
            schema=request_spark_schema,
        )

        # --- Broadcast session + config ---------------------------------
        bc_session = spark_session.sparkContext.broadcast(serialized_session)
        bc_config = spark_session.sparkContext.broadcast(
            pickle.dumps(
                config or SendManyConfig(
                    wait=wait,
                    raise_error=raise_error,
                    normalize=normalize,
                    stream=stream,
                    cache=cache,
                    cache_by=cache_by,
                    cache_anonymize=cache_anonymize,
                    received_from=received_from,
                    received_to=received_to,
                    wait_cache=wait_cache,
                    batch_size=batch_size,
                    ordered=ordered,
                    max_in_flight=max_in_flight,
                )
            )
        )

        # --- Resolve output Arrow schema --------------------------------
        parse_body: bool

        if parse is None or parse is False:
            arrow_schema = RESPONSE_ARROW_SCHEMA
            parse_body = False
        elif parse is True:
            if not request_list:
                raise ValueError(
                    "spark_send(parse=True) requires at least one request "
                    "to discover the output schema, but the request "
                    "iterator was empty."
                )
            probe_resp = self.send(
                request_list[0],
                wait=wait,
                raise_error=raise_error,
                stream=stream,
            )
            if apply is not None:
                probe_resp = apply(probe_resp)
            arrow_schema = probe_resp.to_arrow_table(parse=True).schema
            parse_body = True
        else:
            # parse is a CastOptionsArg
            if isinstance(parse, pa.Schema):
                arrow_schema = parse
            else:
                opts = CastOptions.check_arg(parse)
                arrow_schema = opts.target_arrow_schema

            if arrow_schema is None:
                arrow_schema = RESPONSE_ARROW_SCHEMA
                parse_body = False
            else:
                parse_body = True

        spark_schema = arrow_schema_to_spark_schema(arrow_schema)

        # --- Broadcast helpers ------------------------------------------
        bc_apply = (
            spark_session.sparkContext.broadcast(pickle.dumps(apply))
            if apply is not None else None
        )
        bc_parse_body = spark_session.sparkContext.broadcast(parse_body)

        # --- mapInArrow -------------------------------------------------
        def _map_arrow(batch_iter: Iterator[pa.RecordBatch]) -> Iterator[pa.RecordBatch]:
            """Reconstruct requests from Arrow, send, yield response batches."""
            import yggdrasil.pickle.ser as _pickle

            _session: Session = _pickle.loads(bc_session.value)
            _cfg: SendManyConfig = _pickle.loads(bc_config.value)
            _apply_fn: Optional[Callable] = (
                _pickle.loads(bc_apply.value) if bc_apply is not None else None
            )
            _parse = bc_parse_body.value

            for batch in batch_iter:
                for _req in PreparedRequest.from_arrow(batch):
                    _resp: Response = next(iter(
                        _session.send_many([_req], config=_cfg)
                    ))
                    if _apply_fn is not None:
                        _resp = _apply_fn(_resp)
                    if _parse:
                        yield _resp.to_arrow_table(parse=True).to_batches()[0]
                    else:
                        yield _resp.to_arrow_batch(parse=False)

        return request_df.mapInArrow(_map_arrow, schema=spark_schema)





"""
yggdrasil/dataclasses/expiring.py

Thread-safe expiring cache primitives.

Classes
-------
Expiring[T]
    Abstract base: subclasses implement ``_refresh()`` to reload a single value
    on demand.

ExpiringDict[K, V]
    Thread-safe dict where every key has an individual TTL.  No subclassing
    required — pass a ``refresher`` callable if per-key auto-refresh is wanted.

Change log vs previous version
- ``Expiring``: no ``new_instance`` callable field; subclasses implement
  ``_refresh()``.
- ``ExpiringDict``: new; built on the same time-utils / lock discipline as
  ``Expiring``.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from threading import RLock
from time import time_ns
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

__all__ = [
    # single-value cache
    "Expiring",
    "ExpiredError",
    "RefreshResult",
    # dict cache
    "ExpiringDict",
    # time utils
    "now_utc_ns",
    "datetime_to_epoch_ns",
    "timedelta_to_ns",
]

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
_IntLike = Union[int, str]

_MISSING = object()


# ═══════════════════════════════════════════════════════════
# Time utilities  (shared by both Expiring and ExpiringDict)
# ═══════════════════════════════════════════════════════════

def now_utc_ns() -> int:
    """UTC epoch nanoseconds (monotonic-wall hybrid via time_ns)."""
    return time_ns()


_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


def _coerce_utc(dt: datetime) -> datetime:
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def datetime_to_epoch_ns(dt: datetime) -> int:
    """datetime → epoch ns (microsecond precision ⇒ trailing *000)."""
    dt_utc = _coerce_utc(dt)
    delta = dt_utc - _EPOCH
    return (
        delta.days * 86_400 * 1_000_000_000
        + delta.seconds * 1_000_000_000
        + delta.microseconds * 1_000
    )


def timedelta_to_ns(td: timedelta) -> int:
    """timedelta → ns (microsecond precision ⇒ trailing *000)."""
    return (
        td.days * 86_400 * 1_000_000_000
        + td.seconds * 1_000_000_000
        + td.microseconds * 1_000
    )


# ── casting helpers ──────────────────────────────────────────

def _to_intlike(x: Any, name: str) -> int:
    try:
        return int(x)
    except Exception as e:
        raise TypeError(f"{name} must be int-like, got {type(x).__name__}") from e


def _to_epoch_ns(x: Any, name: str) -> int:
    if isinstance(x, datetime):
        return datetime_to_epoch_ns(x)
    return _to_intlike(x, name)


def _to_ttl_ns(x: Any) -> Optional[int]:
    if x is None:
        return None
    if isinstance(x, timedelta):
        return timedelta_to_ns(x)
    return _to_intlike(x, "ttl_ns")


def _to_expires_at_ns(x: Any, *, now_ns: int) -> Optional[int]:
    """
    Convert an expiry input to an absolute epoch-ns timestamp.

    Accepts:
    - ``None`` → no expiry
    - ``int``-like absolute epoch ns
    - ``datetime`` absolute time
    - ``timedelta`` → now + delta
    """
    if x is None:
        return None
    if isinstance(x, timedelta):
        return now_ns + timedelta_to_ns(x)
    if isinstance(x, datetime):
        return datetime_to_epoch_ns(x)
    return _to_intlike(x, "expires_at_ns")


def _ttl_ns_to_seconds(ttl_ns: Optional[int]) -> Optional[float]:
    return None if ttl_ns is None else ttl_ns / 1_000_000_000


def _seconds_to_ns(seconds: Optional[float]) -> Optional[int]:
    return None if seconds is None else int(seconds * 1_000_000_000)


# ═══════════════════════════════════════════════════════════
# Expiring[T]  — single-value abstract cache
# ═══════════════════════════════════════════════════════════

class ExpiredError(RuntimeError):
    pass


@dataclass(slots=True, frozen=True)
class RefreshResult(Generic[T]):
    """
    Output of ``_refresh()``.

    Provide either:
    - ``ttl_ns`` (duration) OR
    - ``expires_at_ns`` (absolute time)

    If both are ``None`` → non-expiring.
    """
    value: Optional[T]
    created_at_ns: Optional[int] = None
    ttl_ns: Optional[int] = None
    expires_at_ns: Optional[int] = None

    @classmethod
    def make(
        cls,
        value: Optional[T] = None,
        *,
        created_at_ns: Optional[_IntLike] = None,
        ttl_ns: Optional[Union[_IntLike, timedelta]] = None,
        expires_at_ns: Optional[_IntLike] = None,
    ) -> "RefreshResult[T]":
        """Convenience constructor with defaults + light casting."""
        created = None if created_at_ns is None else _to_intlike(created_at_ns, "created_at_ns")
        ttl = _to_ttl_ns(ttl_ns)
        exp = None if expires_at_ns is None else _to_intlike(expires_at_ns, "expires_at_ns")
        return cls(value=value, created_at_ns=created, ttl_ns=ttl, expires_at_ns=exp)


@dataclass(slots=True)
class Expiring(Generic[T], ABC):
    """
    Abstract thread-safe expiring cache holder.

    Subclasses implement ``_refresh()`` which is called when ``.value`` is
    accessed and the cache is expired.
    """

    _value: Optional[T] = None
    _created_at_ns: int = field(default_factory=now_utc_ns, repr=False)
    _expires_at_ns: Optional[int] = field(default=None, repr=False)
    _ttl_ns: Optional[int] = field(default=None, repr=False)
    _lock: RLock = field(default_factory=RLock, init=False, repr=False, compare=False)

    # ── abstract ────────────────────────────────────────────

    @abstractmethod
    def _refresh(self) -> RefreshResult[T]:
        """
        Fetch a fresh value and return refresh metadata.

        Called **outside** the internal lock.
        Must not mutate ``self`` directly — return a ``RefreshResult`` instead.
        """

    # ── constructors ────────────────────────────────────────

    @classmethod
    def create(
        cls,
        value: Optional[Union[T, RefreshResult[T]]] = None,
        *args,
        ttl: Optional[Union[_IntLike, timedelta]] = None,
        expires_at: Optional[Union[_IntLike, datetime, timedelta]] = None,
        created_at: Optional[Union[_IntLike, datetime]] = None,
        **kwargs,
    ) -> "Expiring[T]":
        if isinstance(value, RefreshResult):
            created_at = value.created_at_ns
            exp_val = value.expires_at_ns
            ttl_val = value.ttl_ns
        else:
            created_at = _to_epoch_ns(
                created_at if created_at is not None else now_utc_ns(), "now_ns"
            )
            ttl_val = _to_ttl_ns(ttl)
            exp_val = _to_expires_at_ns(expires_at, now_ns=created_at)

        if exp_val is None and created_at is not None and ttl_val is not None:
            exp_val = created_at + ttl_val

        return cls(  # type: ignore[misc]
            _value=value,
            _created_at_ns=created_at,
            _expires_at_ns=exp_val,
            _ttl_ns=ttl_val,
            *args,
            **kwargs,
        )

    # ── value property ───────────────────────────────────────

    @property
    def value(self) -> Optional[T]:
        """Cached value with auto-refresh on expiry."""
        if self._value is None:
            result = self._refresh()
            self._apply_refresh_result(result)
            with self._lock:
                return self._value

        now_val = now_utc_ns()
        with self._lock:
            if not self._is_expired_locked(now_val):
                return self._value

        result = self._refresh()
        self._apply_refresh_result(result)
        with self._lock:
            return self._value

    @value.setter
    def value(self, v: Optional[T]) -> None:
        with self._lock:
            self._value = v

    # ── tuning properties ────────────────────────────────────

    @property
    def ttl_ns(self) -> Optional[int]:
        with self._lock:
            return self._ttl_ns

    @ttl_ns.setter
    def ttl_ns(self, v: Any) -> None:
        ttl_val = _to_ttl_ns(v)
        with self._lock:
            self._ttl_ns = ttl_val
            if ttl_val is None:
                return
            self._expires_at_ns = self._created_at_ns + ttl_val

    @property
    def expires_at_ns(self) -> Optional[int]:
        with self._lock:
            return self._expires_at_ns

    @expires_at_ns.setter
    def expires_at_ns(self, v: Any) -> None:
        with self._lock:
            self._expires_at_ns = _to_expires_at_ns(v, now_ns=now_utc_ns())

    # ── convenience ──────────────────────────────────────────

    def is_expired(self, *, now_ns: Optional[Union[_IntLike, datetime]] = None) -> bool:
        now_val = _to_epoch_ns(now_ns if now_ns is not None else now_utc_ns(), "now_ns")
        with self._lock:
            return self._is_expired_locked(now_val)

    def refresh(self) -> None:
        """Force refresh now (same as accessing ``.value`` and discarding it)."""
        _ = self.value

    # ── pickle ───────────────────────────────────────────────

    def __getstate__(self) -> dict[str, Any]:
        with self._lock:
            return {
                "_value": self._value,
                "_created_at_ns": self._created_at_ns,
                "_expires_at_ns": self._expires_at_ns,
                "_ttl_ns": self._ttl_ns,
            }

    def __setstate__(self, state: dict[str, Any]) -> None:
        object.__setattr__(self, "_lock", RLock())
        with self._lock:
            self._value = state.get("_value")
            self._created_at_ns = _to_epoch_ns(
                state.get("_created_at_ns", now_utc_ns()), "_created_at_ns"
            )
            self._ttl_ns = _to_ttl_ns(state.get("_ttl_ns"))
            self._expires_at_ns = _to_expires_at_ns(
                state.get("_expires_at_ns"), now_ns=self._created_at_ns
            )

    # ── internals ────────────────────────────────────────────

    def _apply_refresh_result(self, rr: RefreshResult[T]) -> None:
        created = int(rr.created_at_ns) if rr.created_at_ns is not None else now_utc_ns()
        ttl = int(rr.ttl_ns) if rr.ttl_ns is not None else None
        exp = int(rr.expires_at_ns) if rr.expires_at_ns is not None else None
        if exp is None and ttl is not None:
            exp = created + ttl
        with self._lock:
            self._value = rr.value
            self._created_at_ns = created
            self._ttl_ns = ttl
            self._expires_at_ns = exp

    def _effective_expires_at_ns_locked(self) -> Optional[int]:
        if self._expires_at_ns is not None:
            return int(self._expires_at_ns)
        if self._ttl_ns is None:
            return None
        return int(self._created_at_ns) + int(self._ttl_ns)

    def _is_expired_locked(self, now_val: int) -> bool:
        exp = self._effective_expires_at_ns_locked()
        return exp is not None and now_val >= exp


# ═══════════════════════════════════════════════════════════
# ExpiringDict[K, V]  — per-key TTL dict
# ═══════════════════════════════════════════════════════════

class ExpiringDict(Generic[K, V]):
    """
    Thread-safe dictionary where every key carries an individual TTL.

    Built on the same nanosecond time-utils and ``RLock`` discipline as
    ``Expiring``.  No subclassing required.

    Parameters
    ----------
    default_ttl :
        Default TTL as seconds (``float``), nanoseconds (``int``),
        ``timedelta``, or ``None`` (keys never expire unless given a per-key
        TTL).
    max_size :
        Evict the soonest-to-expire key when capacity is reached.
    refresher :
        Optional ``Callable[[key], RefreshResult[V]]``.  When supplied,
        an expired get will call ``refresher(key)`` and atomically replace the
        entry rather than returning the default/raising.

    Serialization
    -------------
    ``__getstate__`` / ``__setstate__`` are implemented: only live (non-expired)
    entries are persisted; the lock is recreated on load.  Compatible with
    ``pickle``, ``copy.deepcopy``, and ``joblib``.
    """

    # ── construction ─────────────────────────────────────────

    def __init__(
        self,
        default_ttl: Optional[Union[float, int, timedelta]] = 300.0,
        *,
        max_size: Optional[int] = None,
        refresher: Optional[Callable[[K], RefreshResult[V]]] = None,
    ) -> None:
        self._default_ttl_ns: Optional[int] = self._parse_ttl(default_ttl)
        self._max_size = max_size
        self._refresher = refresher
        # {key: (value, expires_at_ns)}  — expires_at_ns is epoch ns or None
        self._store: Dict[K, Tuple[V, Optional[int]]] = {}
        self._lock: RLock = RLock()

    # ── serialization ────────────────────────────────────────

    def __getstate__(self) -> Dict[str, Any]:
        with self._lock:
            now = now_utc_ns()
            return {
                "default_ttl_ns": self._default_ttl_ns,
                "max_size": self._max_size,
                # refresher is intentionally excluded (not picklable in general)
                "store": {
                    k: (v, exp)
                    for k, (v, exp) in self._store.items()
                    if exp is None or now < exp
                },
            }

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self._default_ttl_ns = state.get("default_ttl_ns")
        self._max_size = state.get("max_size")
        self._refresher = None
        self._store = state.get("store", {})
        self._lock = RLock()

    # ── core helpers ─────────────────────────────────────────

    @staticmethod
    def _parse_ttl(ttl: Any) -> Optional[int]:
        """Convert seconds/timedelta/int-ns/None → nanoseconds int or None."""
        if ttl is None:
            return None
        if isinstance(ttl, timedelta):
            return timedelta_to_ns(ttl)
        if isinstance(ttl, float):
            return _seconds_to_ns(ttl)
        return int(ttl)  # treat bare int as nanoseconds for internal use

    def _expires_at_ns(self, ttl: Any) -> Optional[int]:
        """
        Compute absolute expiry from a per-call TTL override or fall back to
        the instance default.  Returns ``None`` if no TTL is set.
        """
        if ttl is _MISSING:
            ns = self._default_ttl_ns
        else:
            ns = self._parse_ttl(ttl)
        return None if ns is None else now_utc_ns() + ns

    def _is_expired(self, expires_at: Optional[int]) -> bool:
        return expires_at is not None and now_utc_ns() >= expires_at

    def _evict_expired_locked(self) -> None:
        now = now_utc_ns()
        dead = [k for k, (_, exp) in self._store.items() if exp is not None and now >= exp]
        for k in dead:
            del self._store[k]

    def _evict_one_for_capacity_locked(self) -> None:
        if not self._store:
            return
        # Prefer keys with the soonest expiry; treat None (non-expiring) as ∞
        victim = min(
            self._store,
            key=lambda k: self._store[k][1] if self._store[k][1] is not None else float("inf"),
        )
        del self._store[victim]

    def _put_locked(self, key: K, value: V, expires_at: Optional[int]) -> None:
        """Write a key into the store, respecting max_size."""
        if (
            self._max_size is not None
            and key not in self._store
            and len(self._store) >= self._max_size
        ):
            self._evict_one_for_capacity_locked()
        self._store[key] = (value, expires_at)

    # ── dict-style interface ──────────────────────────────────

    def set(
        self,
        key: K,
        value: V,
        ttl: Any = _MISSING,
    ) -> None:
        """
        Insert or overwrite *key*.

        ``ttl`` accepts seconds (``float``), nanoseconds (``int``),
        ``timedelta``, or ``None`` (no expiry).  Omitting ``ttl`` uses the
        instance default.
        """
        exp = self._expires_at_ns(ttl)
        with self._lock:
            self._evict_expired_locked()
            self._put_locked(key, value, exp)

    def __setitem__(self, key: K, value: V) -> None:
        self.set(key, value)

    def get(self, key: K, default: Any = None) -> Optional[V]:
        """Return value for *key*, or *default* if missing / expired."""
        with self._lock:
            entry = self._store.get(key)
            if entry is None:
                return default
            value, expires_at = entry
            if not self._is_expired(expires_at):
                return value
            del self._store[key]

        # Expired — try refresher if configured
        if self._refresher is not None:
            try:
                rr = self._refresher(key)
            except Exception:
                return default
            exp = rr.expires_at_ns
            if exp is None and rr.ttl_ns is not None:
                created = rr.created_at_ns or now_utc_ns()
                exp = created + rr.ttl_ns
            with self._lock:
                self._put_locked(key, rr.value, exp)
            return rr.value

        return default

    def __getitem__(self, key: K) -> V:
        val = self.get(key, _MISSING)
        if val is _MISSING:
            raise KeyError(key)
        return val  # type: ignore[return-value]

    def __delitem__(self, key: K) -> None:
        with self._lock:
            if key not in self._store:
                raise KeyError(key)
            del self._store[key]

    def pop(self, key: K, *args) -> V:
        with self._lock:
            entry = self._store.pop(key, _MISSING)
        if entry is _MISSING:
            if args:
                return args[0]
            raise KeyError(key)
        value, expires_at = entry
        if self._is_expired(expires_at):
            return args[0] if args else None  # type: ignore[return-value]
        return value

    def __contains__(self, key: object) -> bool:
        return self.get(key, _MISSING) is not _MISSING  # type: ignore[arg-type]

    def __len__(self) -> int:
        with self._lock:
            self._evict_expired_locked()
            return len(self._store)

    def __iter__(self) -> Iterator[K]:
        return iter(self.keys())

    def __repr__(self) -> str:
        default_s = (
            f"{self._default_ttl_ns / 1e9:.3g}s"
            if self._default_ttl_ns is not None
            else "∞"
        )
        return f"{self.__class__.__name__}(len={len(self)}, default_ttl={default_s})"

    # ── bulk operations ───────────────────────────────────────

    def set_many(
        self,
        mapping: Dict[K, V],
        ttl: Any = _MISSING,
    ) -> None:
        """Atomically insert multiple key-value pairs sharing a TTL."""
        exp = self._expires_at_ns(ttl)
        with self._lock:
            self._evict_expired_locked()
            for key, value in mapping.items():
                self._put_locked(key, value, exp)

    def update(
        self,
        other: Union[Dict[K, V], "ExpiringDict[K, V]", None] = None,
        ttl: Any = _MISSING,
        **kwargs: V,
    ) -> None:
        """
        Update the dict from a mapping and/or keyword arguments, mirroring
        the stdlib ``dict.update`` signature.

        Parameters
        ----------
        other :
            A plain ``dict``, another ``ExpiringDict``, or any object with an
            ``.items()`` method.  When the source is an ``ExpiringDict`` and no
            explicit *ttl* is given, each key's **remaining TTL is preserved**
            so expiry semantics survive a copy.  Already-expired source keys
            are silently skipped.
        ttl :
            Override TTL for every key written.  Accepts seconds (``float``),
            nanoseconds (``int``), ``timedelta``, or ``None`` (no expiry).
            When omitted, the instance default is used for plain-dict sources,
            or the source key's remaining TTL is used for ``ExpiringDict``
            sources.
        **kwargs :
            Additional key-value pairs merged after *other*, using *ttl* /
            the instance default.
        """
        explicit_ttl = ttl is not _MISSING

        # Snapshot source outside our lock to avoid lock-ordering deadlocks
        if isinstance(other, ExpiringDict):
            snapshot_ts = now_utc_ns()
            with other._lock:
                source_raw: list = list(other._store.items())  # [(key, (value, exp))]
            is_expiring_dict = True
        elif other is not None:
            if hasattr(other, "items"):
                source_raw = list(other.items())
            else:
                source_raw = list(other)
            is_expiring_dict = False
        else:
            source_raw = []
            is_expiring_dict = False

        with self._lock:
            self._evict_expired_locked()

            for key, raw in source_raw:
                if is_expiring_dict:
                    value, src_exp = raw
                    # Skip source keys that had already expired at snapshot time
                    if src_exp is not None and snapshot_ts >= src_exp:
                        continue
                    if explicit_ttl:
                        exp = self._expires_at_ns(ttl)
                    elif src_exp is None:
                        # Source key is non-expiring; apply instance default
                        exp = self._expires_at_ns(_MISSING)
                    else:
                        # Carry over the remaining duration faithfully
                        remaining_ns = src_exp - snapshot_ts
                        exp = now_utc_ns() + remaining_ns
                else:
                    value, exp = raw, self._expires_at_ns(ttl)

                self._put_locked(key, value, exp)

            for key, value in kwargs.items():
                self._put_locked(key, value, self._expires_at_ns(ttl))

    def get_many(self, keys: Iterable[K]) -> Dict[K, V]:
        """Return ``{key: value}`` for all live keys in *keys*."""
        result: Dict[K, V] = {}
        now = now_utc_ns()
        with self._lock:
            for key in keys:
                entry = self._store.get(key)
                if entry is not None:
                    value, exp = entry
                    if exp is None or now < exp:
                        result[key] = value
                    else:
                        del self._store[key]
        return result

    def delete_many(self, keys: Iterable[K]) -> int:
        """Delete *keys*; returns count of keys actually removed."""
        removed = 0
        with self._lock:
            for key in keys:
                if self._store.pop(key, _MISSING) is not _MISSING:
                    removed += 1
        return removed

    # ── inspection ───────────────────────────────────────────

    def ttl(self, key: K) -> Optional[float]:
        """Remaining TTL in **seconds** for *key*, or ``None`` if gone/expired."""
        with self._lock:
            entry = self._store.get(key)
            if entry is None:
                return None
            _, expires_at = entry
            if expires_at is None:
                return None  # non-expiring
            remaining_ns = expires_at - now_utc_ns()
            if remaining_ns <= 0:
                del self._store[key]
                return None
            return remaining_ns / 1_000_000_000

    def ttl_ns(self, key: K) -> Optional[int]:
        """Remaining TTL in **nanoseconds** for *key*, or ``None``."""
        with self._lock:
            entry = self._store.get(key)
            if entry is None:
                return None
            _, expires_at = entry
            if expires_at is None:
                return None
            remaining = expires_at - now_utc_ns()
            if remaining <= 0:
                del self._store[key]
                return None
            return remaining

    def keys(self) -> list[K]:
        now = now_utc_ns()
        with self._lock:
            return [k for k, (_, exp) in self._store.items() if exp is None or now < exp]

    def values(self) -> list[V]:
        now = now_utc_ns()
        with self._lock:
            return [v for v, exp in self._store.values() if exp is None or now < exp]

    def items(self) -> list[Tuple[K, V]]:
        now = now_utc_ns()
        with self._lock:
            return [(k, v) for k, (v, exp) in self._store.items() if exp is None or now < exp]

    def snapshot(self) -> Dict[K, Tuple[V, Optional[int]]]:
        """
        Shallow copy of live entries as ``{key: (value, expires_at_ns)}``.
        Useful for debugging or external persistence.
        """
        now = now_utc_ns()
        with self._lock:
            return {
                k: (v, exp)
                for k, (v, exp) in self._store.items()
                if exp is None or now < exp
            }

    def clear(self) -> None:
        with self._lock:
            self._store.clear()

    def purge_expired(self) -> int:
        """Explicitly evict all expired keys; returns count removed."""
        with self._lock:
            before = len(self._store)
            self._evict_expired_locked()
            return before - len(self._store)

    # ── refresh / touch ───────────────────────────────────────

    def refresh_key(self, key: K, ttl: Any = _MISSING) -> bool:
        """
        Reset the TTL of an existing, live key.
        Returns ``True`` if key existed and was refreshed, ``False`` otherwise.
        """
        exp = self._expires_at_ns(ttl)
        with self._lock:
            entry = self._store.get(key)
            if entry is None:
                return False
            value, old_exp = entry
            if self._is_expired(old_exp):
                del self._store[key]
                return False
            self._store[key] = (value, exp)
            return True

    def get_or_set(
        self,
        key: K,
        default: Union[V, Callable[[], V]],
        ttl: Any = _MISSING,
    ) -> V:
        """
        Return the live value for *key*; if missing/expired, store *default*
        (or the result of calling it) and return that.
        """
        with self._lock:
            entry = self._store.get(key)
            if entry is not None:
                value, expires_at = entry
                if not self._is_expired(expires_at):
                    return value
            value = default() if callable(default) else default
            exp = self._expires_at_ns(ttl)
            self._put_locked(key, value, exp)
            return value

    # ── RefreshResult convenience ─────────────────────────────

    def apply_refresh_result(self, key: K, rr: RefreshResult[V]) -> None:
        """
        Store the outcome of an external ``RefreshResult`` for *key*.
        Mirrors the contract from ``Expiring._apply_refresh_result``.
        """
        created = int(rr.created_at_ns) if rr.created_at_ns is not None else now_utc_ns()
        ttl_ns_val = int(rr.ttl_ns) if rr.ttl_ns is not None else None
        exp = int(rr.expires_at_ns) if rr.expires_at_ns is not None else None
        if exp is None and ttl_ns_val is not None:
            exp = created + ttl_ns_val
        with self._lock:
            self._put_locked(key, rr.value, exp)